"""
File contains:

    - :class:`PhysNode`
    - :class:`PhysTree`

Author: W. Wybo
"""

import numpy as np

import warnings

from . import morphtree
from .morphtree import MorphNode, MorphTree
from ..channels import concmechs, ionchannels


class PhysNode(MorphNode):
    """
    Node associated with `neat.PhysTree`. Stores the physiological parameters
    of the cylindrical segment connecting this node with its parent node

    Attributes
    ----------
    currents: dict {str: [float,float]}
        dict with as keys the channel names and as values lists of length two
        containing as first entry the channels' conductance density (uS/cm^2)
        and as second element the channels reversal (mV) (i.e.:
        {name: [g_max (uS/cm^2), e_rev (mV)]})
        For the leak conductance, the corresponding key is 'L'
    concmechs: dict
        dict containing concentration mechanisms present in the segment
    c_m: float
        The sement's specific membrane capacitance (uF/cm^2)
    r_a: float
        The segment's axial resistance (MOhm*cm)
    g_shunt: float
        Point-like shunt conductance located at x=1 (uS)
    e_eq: float
        Segment's equilibrium potential
    """
    def __init__(self, index, p3d=None,
                       c_m=1., r_a=100*1e-6, g_shunt=0., e_eq=-75.):
        super().__init__(index, p3d)
        self.currents = {} #{name: (g_max (uS/cm^2), e_rev (mV))}
        self.concmechs = {}
        # biophysical parameters
        self.c_m = c_m # uF/cm^2
        self.r_a = r_a # MOhm*cm
        self.g_shunt = g_shunt # uS
        self.e_eq = e_eq

    def setPhysiology(self, c_m, r_a, g_shunt=0.):
        """
        Set the physiological parameters of the current

        Parameters
        ----------
        c_m: float
            the membrance capacitance (uF/cm^2)
        r_a: float
            the axial current (MOhm*cm)
        g_shunt: float
            A point-like shunt, located at x=1 on the node. Defaults to 0.
        """
        self.c_m = c_m # uF/cm^2
        self.r_a = r_a # MOhm*cm
        self.g_shunt = g_shunt

    def _addCurrent(self, channel_name, g_max, e_rev):
        """
        Add an ion channel current at this node. ('L' as `channel_name`
        signifies the leak current)

        Parameters
        ----------
        channel_name: string
            the name of the current
        g_max: float
            the conductance of the current (uS/cm^2)
        e_rev: float
            the reversal potential of the current (mV)
        """
        self.currents[channel_name] = [g_max, e_rev]

    def addConcMech(self, ion, params={}):
        """
        Add a concentration mechanism at this node.

        Parameters
        ----------
        ion: string
            the ion the mechanism is for
        params: dict
            parameters for the concentration mechanism (only used for NEURON model)
        """
        if set(params.keys()) == {'gamma', 'tau'}:
            self.concmechs[ion] = concmechs.ExpConcMech(ion,
                                        params['tau'], params['gamma'])
        else:
            warnings.warn('These parameters do not match any NEAT concentration ' + \
                          'mechanism, no concentration mechanism has been added', UserWarning)

    def setEEq(self, e_eq):
        """
        Set the equilibrium potential at the node.

        Parameters
        ----------
        e_eq: float
            the equilibrium potential (mV)
        """
        self.e_eq = e_eq

    def fitLeakCurrent(self, channel_storage, e_eq_target=-75., tau_m_target=10.):
        """
        """
        gsum = 0.
        i_eq = 0.

        for channel_name in set(self.currents.keys()) - set('L'):
            g, e = self.currents[channel_name]
            # get the ionchannel object
            channel = channel_storage[channel_name]
            # compute channel conductance and current
            p_open = channel.computePOpen(e_eq_target)
            g_chan = g * p_open
            i_chan = g_chan * (e - e_eq_target)
            gsum += g_chan
            i_eq += i_chan

        if self.c_m / (tau_m_target*1e-3) < gsum:
            warnings.warn('Membrane time scale is chosen larger than ' + \
                          'possible, adding small leak conductance')
            tau_m_target = self.c_m / (gsum + 20.)
        else:
            tau_m_target *= 1e-3
        g_l = self.c_m / tau_m_target - gsum
        e_l = e_eq_target - i_eq / g_l
        self.currents['L'] = [g_l, e_l]
        self.e_eq = e_eq_target

    def getGTot(self, channel_storage, v=None):
        """
        Get the total conductance of the membrane at a steady state given voltage,
        if nothing is given, the equilibrium potential is used to compute membrane
        conductance.

        Parameters
        ----------
            channel_storage: dict {``channel_name``: `channel_instance`}
                dict where all ion channel objects present on the node are stored
            v: float (optional, defaults to `self.e_eq`)
                the potential (in mV) at which to compute the membrane conductance

        Returns
        -------
            float
                the total conductance of the membrane (uS / cm^2)
        """
        v = self.e_eq if v is None else v
        g_tot = self.currents['L'][0]
        for channel_name in set(self.currents.keys()) - set('L'):
            g, e = self.currents[channel_name]
            # get the ionchannel object
            channel = channel_storage[channel_name]
            g_tot += g * channel.computePOpen(v)

        return g_tot

    def asPassiveMembrane(self, channel_storage, v=None):
        v = self.e_eq if v is None else v
        g_l = self.getGTot(channel_storage, v=v)
        t_m = self.c_m / g_l * 1e3 # time scale in ms
        self.currents = {'L': (0., 0.)} # dummy values
        self.fitLeakCurrent(channel_storage, e_eq_target=v, tau_m_target=t_m)

    def __str__(self, with_parent=False, with_children=False):
        node_string = super().__str__()
        if self.parent_node is not None:
            node_string += ', Parent: ' + super().__str__()
        node_string += ' --- (r_a = ' + str(self.r_a) + ' MOhm*cm, ' + \
                       ', '.join(['g_' + cname + ' = ' + str(cpar[0]) + ' uS/cm^2' \
                            for cname, cpar in self.currents.items()]) + \
                       ', c_m = ' + str(self.c_m) + ' uF/cm^2)'
        return node_string


class PhysTree(MorphTree):
    """
    Adds physiological parameters to `neat.MorphTree` and convenience functions
    to set them across the morphology. Initialized in the same way as
    `neat.MorphTree`

    Attributes
    ----------
    channel_storage: dict {str: `neat.IonChannel`}
        Stores the user defined ion channels present in the tree
    """
    def __init__(self, file_n=None, types=[1,3,4]):
        super().__init__(file_n=file_n, types=types)
        # set basic physiology parameters (c_m = 1.0 uF/cm^2 and
        # r_a = 0.0001 MOhm*cm)
        for node in self:
            node.setPhysiology(1.0, 100./1e6)
        self.channel_storage = {}

    def _createCorrespondingNode(self, node_index, p3d=None,
                                      c_m=1., r_a=100*1e-6, g_shunt=0., e_eq=-75.):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        return PhysNode(node_index, p3d=p3d)

    @morphtree.originalTreetypeDecorator
    def asPassiveMembrane(self, node_arg=None):
        """
        Makes the membrane act as a passive membrane (for the nodes in
        ``node_arg``), channels are assumed to add a conductance of
        g_max * p_open to the membrane conductance, where p_open for each node
        is evaluated at the equilibrium potential stored in that node

        Parameters
        ----------
        node_arg: optional
                see documentation of :func:`MorphTree._convertNodeArgToNodes`.
                Defaults to None. The nodes for which the membrane is set to
                passive
        """
        for node in self._convertNodeArgToNodes(node_arg):
            node.asPassiveMembrane(self.channel_storage)

    def _distr2Float(self, distr, node, argname=''):
        if isinstance(distr, float):
            val = distr
        elif isinstance(distr, dict):
            val = distr[node.index]
        elif hasattr(distr, '__call__'):
            d2s = self.pathLength({'node': node.index, 'x': .5}, (1., 0.5))
            val = distr(d2s)
        else:
            raise TypeError(argname + ' argument should be a float, dict ' + \
                            'or a callable')
        return val

    @morphtree.originalTreetypeDecorator
    def setEEq(self, e_eq_distr, node_arg=None):
        """
        Set the equilibrium potentials throughout the tree

        Parameters
        ----------
        e_eq_distr: float, dict or :func:`float -> float`
            The equilibrium potentials [mV]
        """
        for node in self._convertNodeArgToNodes(node_arg):
            e = self._distr2Float(e_eq_distr, node, argname='`e_eq_distr`')
            node.setEEq(e)

    @morphtree.originalTreetypeDecorator
    def setPhysiology(self, c_m_distr, r_a_distr, g_s_distr=None, node_arg=None):
        """
        Set specifice membrane capacitance, axial resistance and (optionally)
        static point-like shunt conductances in the tree. Capacitance is stored
        at each node as the attribute 'c_m' (uF/cm2) and axial resistance as the
        attribute 'r_a' (MOhm*cm)

        Parameters
        ----------
        c_m_distr: float, dict or :func:`float -> float`
            specific membrance capacitance
        r_a_distr: float, dict or :func:`float -> float`
            axial resistance
        g_s_distr: float, dict, :func:`float -> float` or None (optional, default
            is `None`)
            point like shunt conductances (placed at `(node.index, 1.)` for the
            nodes in ``node_arg``). By default no shunt conductances are added
        node_arg: optional
            see documentation of :func:`MorphTree._convertNodeArgToNodes`.
            Defaults to None
        """
        for node in self._convertNodeArgToNodes(node_arg):
            c_m = self._distr2Float(c_m_distr, node, argname='`c_m_distr`')
            r_a = self._distr2Float(r_a_distr, node, argname='`r_a_distr`')
            g_s = self._distr2Float(g_s_distr, node, argname='`g_s_distr`') if \
                  g_s_distr is not None else 0.
            node.setPhysiology(c_m, r_a, g_s)

    @morphtree.originalTreetypeDecorator
    def setLeakCurrent(self, g_l_distr, e_l_distr, node_arg=None):
        """
        Set the parameters of the leak current. At each node, leak is stored
        under the attribute `node.currents['L']` at a tuple `(g_l, e_l)` with
        `g_l` the conductance [uS/cm^2] and `e_l` the reversal [mV]

        parameters:
        ----------
        g_l_distr: float, dict or :func:`float -> float`
            If float, the leak conductance is set to this value for all
            the nodes specified in `node_arg`. If it is a function, the input
            must specify the distance from the soma (micron) and the output
            the leak conductance [uS/cm^2] at that distance. If it is a
            dict, keys are the node indices and values the ion leak
            conductances [uS/cm^2].
        e_l_distr: float, dict or :func:`float -> float`
            If float, the reversal [mV] is set to this value for all
            the nodes specified in `node_arg`. If it is a function, the input
            must specify the distance from the soma [um] and the output
            the reversal at that distance. If it is a
            dict, keys are the node indices and values the ion reversals.
        node_arg: optional
            see documentation of :func:`MorphTree._convertNodeArgToNodes`.
            Defaults to None
        """
        for node in self._convertNodeArgToNodes(node_arg):
            g_l = self._distr2Float(g_l_distr, node, argname='`g_l_distr`')
            e_l = self._distr2Float(e_l_distr, node, argname='`e_l_distr`')
            node._addCurrent('L', g_l, e_l)

    @morphtree.originalTreetypeDecorator
    def addCurrent(self, channel, g_max_distr, e_rev_distr, node_arg=None):
        """
        Adds a channel to the morphology. At each node, the channel is stored
        under the attribute `node.currents[channel.__class__.__name__]` as a
        tuple `(g_max, e_rev)` with `g_max` the maximal conductance [uS/cm^2]
        and `e_rev` the reversal [mV]

        Parameters
        ----------
        channel_name: :class:`IonChannel`
            The ion channel
        g_max_distr: float, dict or :func:`float -> float`
            If float, the maximal conductance is set to this value for all
            the nodes specified in `node_arg`. If it is a function, the input
            must specify the distance from the soma (micron) and the output
            the ion channel density (uS/cm^2) at that distance. If it is a
            dict, keys are the node indices and values the ion channel
            densities (uS/cm^2).
        e_rev_distr: float, dict or :func:`float -> float`
            If float, the reversal (mV) is set to this value for all
            the nodes specified in `node_arg`. If it is a function, the input
            must specify the distance from the soma (micron) and the output
            the reversal at that distance. If it is a
            dict, keys are the node indices and values the ion reversals.
        node_arg: optional
            see documentation of :func:`MorphTree._convertNodeArgToNodes`.
            Defaults to None
        """
        if not isinstance(channel, ionchannels.IonChannel):
            raise IOError('`channel` argmument needs to be of class `neat.IonChannel`')

        channel_name = channel.__class__.__name__
        self.channel_storage[channel_name] = channel
        # add the ion channel to the nodes
        for node in self._convertNodeArgToNodes(node_arg):
            g_max = self._distr2Float(g_max_distr, node, argname='`g_max_distr`')
            e_rev = self._distr2Float(e_rev_distr, node, argname='`e_rev_distr`')
            node._addCurrent(channel_name, g_max, e_rev)

    @morphtree.originalTreetypeDecorator
    def getChannelsInTree(self):
        """
        Returns list of strings of all channel names in the tree

        Returns
        -------
        list of string
            the channel names
        """
        return list(self.channel_storage.keys())

    @morphtree.originalTreetypeDecorator
    def addConcMech(self, ion, params={}, node_arg=None):
        """
        Add a concentration mechanism to the tree

        Parameters
        ----------
        ion: string
            the ion the mechanism is for
        params: dict
            parameters for the concentration mechanism (only used for NEURON model)
        node_arg:
            see documentation of :func:`MorphTree._convertNodeArgToNodes`.
            Defaults to None
        """
        for node in self._convertNodeArgToNodes(node_arg):
            node.addConcMech(ion, params=params)

    @morphtree.originalTreetypeDecorator
    def fitLeakCurrent(self, e_eq_target_distr, tau_m_target_distr, node_arg=None):
        """
        Fits the leak current to fix equilibrium potential and membrane time-
        scale.

        !!! Should only be called after all ion channels have been added !!!

        Parameters
        ----------
        e_eq_target_distr: float, dict or :func:`float -> float`
            The target reversal potential (mV). If float, the target reversal is
            set to this value for all the nodes specified in `node_arg`. If it
            is a function, the input must specify the distance from the soma (um)
            and the output the target reversal at that distance. If it is a
            dict, keys are the node indices and values the target reversals
        tau_m_target_distr: float, dict or :func:`float -> float`
            The target membrane time-scale (ms). If float, the target time-scale is
            set to this value for all the nodes specified in `node_arg`. If it
            is a function, the input must specify the distance from the soma (um)
            and the output the target time-scale at that distance. If it is a
            dict, keys are the node indices and values the target time-scales
        node_arg:
            see documentation of :func:`MorphTree._convertNodeArgToNodes`.
            Defaults to None
        """
        for node in self._convertNodeArgToNodes(node_arg):
            e_eq_target = self._distr2Float(e_eq_target_distr, node, argname='`g_max_distr`')
            tau_m_target = self._distr2Float(tau_m_target_distr, node, argname='`e_rev_distr`')
            assert tau_m_target > 0.
            node.fitLeakCurrent(e_eq_target=e_eq_target, tau_m_target=tau_m_target,
                                channel_storage=self.channel_storage)

    def _evaluateCompCriteria(self, node, eps=1e-8, rbool=False):
        """
        Return ``True`` if relative difference in any physiological parameters
        between node and child node is larger than margin ``eps``.

        Overrides the `MorphTree._evaluateCompCriteria()` function called by
        `MorphTree.setCompTree()`.

        Parameters
        ----------
        node: ::class::`MorphNode`
            node that is compared to parent node
        eps: float (optional, default ``1e-8``)
            the margin

        return
        ------
        bool
        """
        rbool = super()._evaluateCompCriteria(node, eps=eps, rbool=rbool)

        if not rbool:
            cnode = node.child_nodes[0]
            rbool = np.abs(node.r_a - cnode.r_a) > eps * np.max([node.r_a, cnode.r_a])
        if not rbool:
            rbool = np.abs(node.c_m - cnode.c_m) > eps * np.max([node.c_m, cnode.c_m])
        if not rbool:
            rbool = set(node.currents.keys()) != set(cnode.currents.keys())
        if not rbool:
            for chan_name, channel in node.currents.items():
                if not rbool:
                    rbool = np.abs(channel[0] - cnode.currents[chan_name][0]) > eps * \
                             np.max([np.abs(channel[0]), np.abs(cnode.currents[chan_name][0])])
                if not rbool:
                    rbool = np.abs(channel[1] - cnode.currents[chan_name][1]) > eps * \
                             np.max([np.abs(channel[1]), np.abs(cnode.currents[chan_name][1])])
        if not rbool:
            rbool = node.g_shunt > 0.001*eps

        return rbool

    # @morphtree.originalTreetypeDecorator
    # def _calcFdMatrix(self, dx=10.):
    #     matdict = {}
    #     locs = [{'node': 1, 'x': 0.}]
    #     # set the first element
    #     soma = self.tree.root
    #     matdict[(0,0)] = 4.0*np.pi*soma.R**2 * soma.G
    #     # recursion
    #     cnodes = root.getChildNodes()[2:]
    #     numel_l = [1]
    #     for cnode in cnodes:
    #         if not is_changenode(cnode):
    #             cnode = find_previous_changenode(cnode)[0]
    #         self._fdMatrixFromRoot(cnode, root, 0, numel_l, locs, matdict, dx=dx)
    #     # create the matrix
    #     FDmat = np.zeros((len(locs), len(locs)))
    #     for ind in matdict:
    #         FDmat[ind] = matdict[ind]

    #     return FDmat, locs # caution, not the reduced locs yet

    # def _fdMatrixFromRoot(self, node, pnode, ibranch, numel_l, locs, matdict, dx=10.*1e-4):
    #     numel = numel_l[0]
    #     # distance between the two nodes and radius of the cylinder
    #     radius *= node.R*1e-4; length *= node.L*1e-4
    #     num = np.around(length/dx)
    #     xvals = np.linspace(0.,1.,max(num+1,2))
    #     dx_ = xvals[1]*length
    #     # set the first element
    #     matdict[(ibranch,numel)] = - np.pi*radius**2 / (node.r_a*dx_)
    #     matdict[(ibranch,ibranch)] += np.pi*radius**2 / (node.r_a*dx_)
    #     matdict[(numel,numel)] = 2.*np.pi*radius**2 / (node.r_a*dx_)
    #     matdict[(numel,ibranch)] = - np.pi*radius**2 / (node.r_a*dx_)
    #     locs.append({'node': node._index, 'x': xvals[1]})
    #     # set the other elements
    #     if len(xvals) > 2:
    #         i = 0; j = 0
    #         if len(xvals) > 3:
    #             for x in xvals[2:-1]:
    #                 j = i+1
    #                 matdict[(numel+i,numel+j)] = - np.pi*radius**2 / (node.r_a*dx_)
    #                 matdict[(numel+j,numel+j)] = 2. * np.pi*radius**2 / (node.r_a*dx_)
    #                                            # + 2.*np.pi*radius*dx_*node.G
    #                 matdict[(numel+j,numel+i)] = - np.pi*radius**2 / (node.r_a*dx_)
    #                 locs.append({'node': node._index, 'x': x})
    #                 i += 1
    #         # set the last element
    #         j = i+1
    #         matdict[(numel+i,numel+j)] = - np.pi*radius**2 / (node.r_a*dx_)
    #         matdict[(numel+j,numel+j)] = np.pi*radius**2 / (node.r_a*dx_)
    #         matdict[(numel+j,numel+i)] = - np.pi*radius**2 / (node.r_a*dx_)
    #         locs.append({'node': node._index, 'x': 1.})
    #     numel_l[0] = numel+len(xvals)-1
    #     # assert numel_l[0] == len(locs)
    #     # if node is leaf, then implement other bc
    #     if len(xvals) > 2:
    #         ibranch = numel+j
    #     else:
    #         ibranch = numel
    #     # move on the further elements
    #     for cnode in node.child_nodes:
    #         self._fdMatrixFromRoot(cnode, node, ibranch, numel_l, locs, matdict, dx=dx)









import numpy as np
import scipy.signal as sig
import sklearn.isotonic as ski


## histogram segmentation ################################################
class histogramSegmentator:
    def __init__(self, hist):
        """
        Class to segment a histogram as returned by the numpy histogram
        function. (Delon, 2005; 2007)

        input:
            [hist]: output of the numpy histogram function
        """
        self.hist = hist[0]
        self.xmap = hist[1]
        # basic quantities
        self.N = np.sum(hist[0])
        self.L = hist[0].size
        # make the histogram noisy
        self.hist = self.hist + 1e-10*np.random.rand(self.L)
        # frequencies in every interval
        self.r = self.hist / self.N

    def _pool_adjacent_violators(self, i0, i1, increasing=True):
        """
        Obtains an increasing or decreasing Grenander estimator of
        the histogram over the interval [i0,i1]

        input:
            [i0, i1]: ints, (inclusive) limits of the histogram
            [increasing]: boolean, if True, increasing estimator,
                else, decreasing estimator

        output:
            [r_grenander]: numpy array, obtained Grenander estimator
        """
        X = np.arange(len(self.r[i0:i1+1]))
        IR = ski.IsotonicRegression(increasing=increasing)
        IR.fit(X, self.r[i0:i1+1])
        r_grenander = IR.predict(X)
        return r_grenander

    def test_unimodal_hypothesis(self, i0, i1, ip, eps=1., maximum=True):
        """
        Tests the unimodal hypothesis, i.e. increasing hypothesis in
        [i0,p] and decreasing hypothesis over [p,i1]

        input:
            [i0,p,i1]: ints

        output:
            boolean, whether the hypothesis is valid or not
        """
        assert i0 <= ip and ip <= i1
        if maximum:
            # construct increasing Grenander estimator on [i0,ip]
            # and decreasing Grenander astimator on [ip,i1]
            r_incr = self._pool_adjacent_violators(i0, ip, increasing=True)
            r_decr = self._pool_adjacent_violators(ip, i1, increasing=False)
            # test both the increasing and decreasing hypothesis
            b_incr = self.test_hypothesis(self.r[i0:ip+1], r_incr, eps=eps)
            b_decr = self.test_hypothesis(self.r[ip:i1+1], r_decr, eps=eps)
        else:
            # construct decreasing Grenander estimator on [i0,ip]
            # and increasing Grenander astimator on [ip,i1]
            r_decr = self._pool_adjacent_violators(i0, ip, increasing=False)
            r_incr = self._pool_adjacent_violators(ip, i1, increasing=True)
            # test both the increasing and decreasing hypothesis
            b_decr = self.test_hypothesis(self.r[i0:ip+1], r_decr, eps=eps)
            b_incr = self.test_hypothesis(self.r[ip:i1+1], r_incr, eps=eps)
        return b_incr and b_decr

    def test_hypothesis(self, r, p, eps=1.):
        """
        Tests whether the hypothesis that the array of frequencies [r]
        is generated from an underlying probability distribution [p] is
        valid.

        input:
            [r]: numpy array, array of observed frequencies
            [p]: numpy array, array of probabilies

        output:
            boolean, whether the hypothesis is true
        """
        assert len(r) == len(p)
        N = len(p)
        # compute values for all intervals
        rvec = np.zeros(N*(N+1)//2)
        pvec = np.zeros(N*(N+1)//2)
        for j0 in range(N):
            for j1 in range(j0,N):
                k = (N-j0)*j0 + j1
                rvec[k] = np.sum(r[j0:j1+1])
                pvec[k] = np.sum(p[j0:j1+1])
        # stabilize the log
        rvec += 1e-12
        pvec += 1e-12
        # evaluate relative entropy for all intervals
        Hvec = rvec*np.log(rvec/pvec) + (1.-rvec)*np.log((1.-rvec)/(1.-pvec))
        # hypothesis is false if Hvec contains values larger than this
        return not np.max(Hvec > eps*np.log(N*(N+1.)/2.)/self.N)

    def get_initial_partition(self):
        """
        Partitions the histogram by computing all the minima of the array
        """
        s_inds = [0] + sig.argrelmin(self.hist)[0].tolist() + [len(self.hist)-1]
        p_inds = sig.argrelmax(self.hist)[0].tolist()
        if self.hist[0] > self.hist[1]:
            p_inds = [0] + p_inds
        if self.hist[-2] < self.hist[-1]:
            p_inds = p_inds + [len(self.hist)-1]
        return s_inds, p_inds

    def find_unimodal_extremum(self, i0s, i1s, eps=1., maximum=True):
        """
        If there is a maximum (minimum) that obeys the unimodal hypothesis, returns it
        """
        ip = i0s + np.argmax(self.r[i0s:i1s+1])
        found = self.test_unimodal_hypothesis(i0s, i1s, ip, eps=eps, maximum=maximum)
        if found:
            return ip
        else:
            return None

    def partition_fine_to_coarse(self, fix_minima=False, eps=1., pprint=False):
        """
        Fine to coarse partitioning algorithm.

        output:
            [s_inds]: the minima of the partition
            [p_inds]: the maxima of the partition
        """
        # initial hypothesis
        s_inds, p_inds = self.get_initial_partition()
        # find the coarsest possible partition
        k = 0
        n_iter = 0
        while k<10 and len(s_inds) > 2:
            i = np.random.randint(1, len(s_inds)-1)
            i0s = s_inds[i-1]; i1s = s_inds[i+1]
            ip = self.find_unimodal_extremum(i0s, i1s, eps=eps)
            if ip != None:
                del s_inds[i]
                del p_inds[i]
                p_inds[i-1] = ip
                k = 0
            else:
                k += 1
            k+=1
            n_iter += 1
            if pprint: print('>>> iter: ' + str(n_iter) + ' <<<')
        if fix_minima:
            for k in range(len(p_inds) - 1):
                i0p = p_inds[k]
                i1p = p_inds[k+1]
                im = i0p + np.argmin(self.r[i0p:i1p+1])
                s_inds[k+1] = im
        return s_inds, p_inds


## empirical interpolation method ########################################
class empiricalInterpolation:
    def __init__(self, fmat, xgrid, mutrain):
        assert fmat.shape[0] == xgrid.shape[0]
        assert fmat.shape[1] == mutrain.shape[0]
        self.mutrain = mutrain
        self.xgrid = xgrid
        self.fmat = fmat

    def run_EIM(self, eps=1e-2, Mmax=50, pprint=False):
        if pprint: print('>>> running EIM <<<')
        Q = np.zeros((len(self.xgrid),0))
        I_ps = []
        I_mus = []
        E_ms = []
        # residue matrix for initialization
        R = self.fmat
        # get the first parameter index
        Rsup = various.supremum_norm(R, axis=0)
        iMu = np.argmax(Rsup)
        I_mus.append(iMu)
        # get the first residue
        res = R[:,iMu]
        # iterate untill the desired precision is reached
        M = 0; E_m = eps+1
        while M < Mmax and E_m > eps:
            if pprint: print('> iter', M, ' eps =', E_m)
            M += 1
            # get the interpolation point index and add it to the list
            iP = np.argmax(np.abs(res))
            I_ps.append(iP)
            # get the new basis function and add it to the matrix
            rho_new = res/res[iP]
            Q = np.concatenate((Q, rho_new[:,np.newaxis]), 1)
            # new B matrix
            B = Q[I_ps,:]
            # new parameter functions
            G = np.linalg.solve(B, self.fmat[I_ps,:])
            # new interpolant
            P = np.dot(Q, G)
            # residue matrix for all snapshots
            R = self.fmat - P
            # find the new parameter index
            Rsup = various.supremum_norm(R, axis=0)
            iMu = np.argmax(Rsup)
            I_mus.append(iMu)
            # get the new residue
            res = R[:,iMu]
            # get the new error
            E_m = Rsup[iMu]
            E_ms.append(E_m)
        else:
            if pprint: print('> finished, eps =', E_m)

        self.G = G
        self.Q = Q
        self.M = M

    def get_f_EIM(self, return_der=True):
        self.f_rhos = []
        self.df_rhos = []
        self.f_gammas = []
        self.df_gammas = []

        for rho in self.Q.T:
            # spline interpolation, get the knots
            tck = intp.splrep(self.xgrid, rho, k=3)
            # construct a spline evaluation function and its derivative
            def f(x, tcktuple=tck):
                return intp.splev(x, tcktuple, der=0)
            def df(x, tcktuple=tck):
                return intp.splev(x, tcktuple, der=1)
            self.f_rhos.append(f)
            self.df_rhos.append(df)

        for gamma in self.G:
            # spline interpolation, get the knots
            tck = intp.splrep(self.mutrain, gamma, k=3)
            # construct a spline evaluation function and its derivative
            def f(x, tcktuple=tck):
                return intp.splev(x, tcktuple, der=0)
            def df(x, tcktuple=tck):
                return intp.splev(x, tcktuple, der=1)
            self.f_gammas.append(f)
            self.df_gammas.append(df)

        if return_der:
            return self.f_rhos, self.f_gammas, self.df_rhos, self.df_gammas
        else:
            return self.f_rhos, self.f_gammas

    def f_intp(self, xv, mv, N=None):
        N = N or self.M
        if N > self.M: N = self.M

        return np.sum([f_rho(xv)[:,np.newaxis] * self.f_gammas[i](mv)[np.newaxis,:] for i, f_rho in enumerate(self.f_rhos[:N])], 0)

    def df_intp(self, xv, mv, N=None):
        N = N or self.M
        if N > self.M: N = self.M

        return np.sum([f_rho(xv)[:,np.newaxis] * self.f_gammas[i](mv)[np.newaxis,:] \
                                            for i, f_rho in enumerate(self.f_rhos[:N])], 0)
##########################################################################


import sympy as sp
import numpy as np


class ConcMech(object):
    def __init__(self, ion):
        self.ion = ion

    def iteritems(self):
        yield None, None


class ExpConcMech(ConcMech):
    def __init__(self, ion, tau, gamma):
        self.tau = tau # ms
        self.gamma = gamma # ms * mM / nA
        super().__init__(ion)

    def iteritems(self):
        yield 'gamma', self.gamma
        yield 'tau', self.tau

    def items(self):
        return [('gamma', self.gamma), ('tau', self.tau)]

    def computeLinear(self, freqs):
        return self.gamma / (freqs + 1e3 / self.tau)

    def computeLin(self, freqs):
        return 1. / (freqs + 1e3 / self.tau)

    def __str__(self):
        return 'tau: %.2f ms, gamma: %.6f (ms/nA)'%(self.tau, self.gamma)



import numpy as np
import dill

import matplotlib.patheffects as patheffects
import matplotlib.patches as patches
import matplotlib.cm as cm
import matplotlib.pyplot as pl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D

from ...trees.morphtree import MorphLoc
from ...trees.phystree import PhysTree
from ...trees.greenstree import GreensTree
from ...trees.sovtree import SOVTree
from ...trees.netree import NET, NETNode, Kernel

from ...tools import kernelextraction as ke

import warnings
import copy
import concurrent.futures
import contextlib
import multiprocessing

try:
    from ...tools.simtools.neuron import neuronmodel as neurm
except ModuleNotFoundError:
    warnings.warn('NEURON not available', UserWarning)

    
def cpu_count(use_hyperthreading=True):
    """
    Return number of available cores.
    Makes use of hypterthreading by default.
    """
    if use_hyperthreading:
        return multiprocessing.cpu_count()
    else:
        return multiprocessing.cpu_count() // 2


def consecutive(inds):
    """
    split a list of ints into consecutive sublists
    """
    return np.split(inds, np.where(np.diff(inds) != 1)[0]+1)


def getExpansionPoints(e_hs, channel, only_e_h=False):
    """
    Returns a list of expansion points around which to compute the impedance
    matrix given a set of holding potentials. If the channel has only one state
    variable, the returned expansion points are at the holding potentials, if
    the channels has two state variables, the returned expansions points are
    are different combinations of the state variable values around the holding
    potentials

    Parameters
    ----------
    e_hs: iterable collection
        The holding potentials around which the expansion points are computed
    channel: `neat.channels.ionchannels.IonChannel`
        The ion channels
    only_e_h: bool
        If True, othe entries in ``sv_hs`` are only at ``e_hs``

    Returns
    -------
    sv_hs: list of dict
        Each entry is a state variable expansion point
    e_hs: list of float
        The holding potentials corresponding to each entry in ``sv_hs``
    """
    e_hs = list(e_hs)
    if len(channel.statevars) == 1 or only_e_h:
        sv_hs = [channel.computeVarinf(e_h) for e_h in e_hs]
    elif len(channel.statevars) == 2:
        # evaluate at the holding potentials
        sv_aux = [channel.computeVarinf(e_h) for e_h in e_hs]

        # check which variable is activation
        sv = channel.computeVarinf(np.array([-43.22, -32.22]))
        sind_act = None
        for ii, svar in enumerate(channel.ordered_statevars):
            if sv[svar][1] > sv[svar][0]:
                sind_act = 'ii' if ii == 0 else 'jj'

        # evaluate at combinations of holding potentials
        sv_hs_extra, e_hs_extra = [], []
        sv_o = channel.ordered_statevars
        for ii, sv_1 in enumerate(sv_aux):
            for jj, sv_2 in enumerate(sv_aux):
                sv_hs_extra.append({str(sv_o[0]): sv_1[sv_o[0]],
                                    str(sv_o[1]): sv_2[sv_o[1]]})
                # follow holding potential of activation state variable
                e_hs_extra.append(eval('e_hs[%s]'%sind_act))

        sv_hs = sv_aux + sv_hs_extra
        e_hs = e_hs + e_hs_extra
    else:
        raise Exception('Method only implemented for channels with two ' + \
                        'or less state variables')

    return sv_hs, e_hs


def asPassiveDendrite(phys_tree, factor_lambda=2., t_calibrate=500.):
        """
        Set the dendrites to be passive compartments. Channel conductances at
        the resting potential are added to passive membrane conductance.

        Parameters
        ----------
        phys_tree: `neat.PhysTree()`
            the neuron model
        factor_lambda: float (optional, defaults to 2.)
            multiplies the numbers of compartments given by the lambda rule (to
            compute resting membrane potential)
        t_calibrate: float (optional, defaults to 500. ms)
            The calibration time for the model (should reach resting potential)

        Returns
        -------
        `neat.PhysTree()`
        """
        dt, t_max = .1, 1.
        # create a biophysical simulation model
        sim_tree = phys_tree.__copy__(new_tree=neurm.NeuronSimTree())
        # compute equilibrium potentials
        sim_tree.initModel(dt=dt, factor_lambda=factor_lambda, t_calibrate=t_calibrate)
        sim_tree.storeLocs([(node.index, .5) for node in phys_tree], 'rec locs')
        res = sim_tree.run(t_max)
        sim_tree.deleteModel()
        v_eqs = [v_m[-1] for v_m in res['v_m']]
        # store the equilbirum potential distribution
        phys_tree.setEEq(v_eqs)
        phys_tree.asPassiveMembrane(node_arg='basal')
        phys_tree.asPassiveMembrane(node_arg='apical')
        phys_tree.setCompTree(eps=1e-2)

        return phys_tree


class FitTreeGF(GreensTree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setName('dont save', '')

    def setExpansionPointsForFit(self, sv_h, e_h):
        """
        Set the holding potentials and expansion points for the fit

        Parameters
        ----------
        sv_hs: dict of {string: np.ndarray}
            Keys are the channel names and values are numpy arrays that contain
            the expansion point for each ion channel
        e_h: float
            the holding potential
        """
        for node in self:
            node.setEEq(e_h)
            for c_name, sv in sv_h.items():
                node.setExpansionPoint(c_name, statevar=sv)

    def setName(self, name, path):
        """
        Set the name and path under which the tree will be stored

        Parameters
        ----------
        name: string
            string based on which name is generated (not equal to actual file
            name)
        path: string
            path where the file is to be located
        """
        self.name = name
        self.path = path

    def setImpedancesInTree(self, many_freqs=False, recompute=False, pprint=False):
        """
        Sets the impedances in the tree.

        Parameters
        ----------
        many_freqs: bool (optional, default is ``False``)
            If ``True``, evaluates the impedances over an array suitable to
            apply the inverse Fourrier transform to obtain temporal kernel
            If ``False``, evaluates at zero frequency
        recompute: bool (optional, default is ``False``)
            Force the impedances to be recomputed
        pprint: bool (optional, default is ``False``)
            Print info
        """
        if pprint:
            print('>>> evaluating impedances with ' + str(list(self.channel_storage.keys())))

        if many_freqs:
            freqs = ke.create_logspace_freqarray()
            suffix = 'allfreqs_'
        else:
            freqs = np.array([0.])
            suffix = ''

        e_h_string = '_eh=%.2f'%(self.root.e_eq)

        # create suffix for state variable expansion point if it is specified
        cname_string = ''
        for c_name, channel in self.channel_storage.items():
            cname_string += '_' + c_name + '_'
            try:
                sv_h = self.root.expansion_points[c_name]
                for svar in channel.ordered_statevars:
                    sv = sv_h[str(svar)]
                    cname_string += str(svar) + '=%.8f'%sv
            except (KeyError, TypeError):
                pass

        file_name = 'GF_' + suffix + self.name + e_h_string + cname_string + '.p'

        # check if impedances already exist
        try:
            # ensure that the tree is recomputed if 'recompute' is true
            if recompute:
                raise IOError
            file = open(self.path + file_name, 'rb')
            tree = dill.load(file)
            self.__dict__.update(tree.__dict__)
            file.close()
            del tree
        except (Exception, IOError, EOFError, KeyError) as err:
            if pprint: print('>>>>>Impedances not stored, calculating...')
            self.setCompTree()
            # set the impedances
            self.setImpedance(freqs, pprint=pprint)

            if not 'dont save' in self.name:
                # store the impedance tree
                file = open(self.path + file_name, 'wb')
                dill.dump(self, file)
                file.close()

    def calcNETSteadyState(self, root_loc=None, dx=5., dz=5.):
        if root_loc is None: root_loc = (1, .5)
        root_loc = MorphLoc(root_loc, self)
        # distribute locs on nodes
        st_nodes = self.gatherNodes(self[root_loc['node']])
        d2s_loc = self.pathLength(root_loc, (1,0.5))
        net_locs = self.distributeLocsOnNodes(d2s=np.arange(d2s_loc, 5000., dx),
                                   node_arg=st_nodes, name='net eval')
        # compute the impedance matrix for net calculation
        z_mat = self.calcImpedanceMatrix('net eval', explicit_method=False)[0]
        # assert np.allclose(z_mat, z_mat_)
        # derive the NET
        net = NET()
        self._addNodeToNET(0., z_mat[0,0], z_mat, np.arange(z_mat.shape[0]), None, net,
                           alpha=1., dz=dz)
        net.setNewLocInds()

        return net, z_mat

    def _addNodeToNET(self, z_min, z_max, z_mat, loc_inds, pnode, net, alpha=1., dz=20.):
        # compute mean impedance of node
        inds = [[]]
        while len(inds[0]) == 0:
            inds = np.where((z_mat > z_min) & (z_mat < z_max))
            z_max += dz
        z_node = np.mean(z_mat[inds])
        # subtract impedances of parent nodes
        gammas = np.array([z_node])
        self._subtractParentKernels(gammas, pnode)
        # add a node to the tree
        node = NETNode(len(net), loc_inds, z_kernel=(np.array([alpha]), gammas))
        if pnode != None:
            net.addNodeWithParent(node, pnode)
        else:
            net.root = node
        # recursion for following nodes
        d_inds = consecutive(np.where(np.diag(z_mat) > z_max)[0])
        for di in d_inds:
            if len(di) > 0:
                self._addNodeToNET(z_max, z_max+dz, z_mat[di,:][:,di], loc_inds[di], node, net,
                                       alpha=alpha, dz=dz)

    def _subtractParentKernels(self, gammas, pnode):
        if pnode != None:
            gammas -= pnode.z_kernel['c']
            self._subtractParentKernels(gammas, pnode.parent_node)


class FitTreeSOV(SOVTree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setName('dont save', '')

    def setName(self, name, path):
        """
        Set the name and path under which the tree will be stored

        Parameters
        ----------
        name: string
            string based on which name is generated (not equal to actual file
            name)
        path: string
            path where the file is to be located
        """
        self.name = name
        self.path = path

    def setSOVInTree(self, recompute=False, pprint=False, maxspace_freq=100.):
        file_name = 'SOV_' + self.name + '.p'
        # load or compute the separation of variables tree
        try:
            # ensure that the tree is recomputed if 'recompute' is true
            if recompute:
                raise IOError
            file = open(self.path + file_name, 'rb')
            tree = dill.load(file)
            self.__dict__.update(tree.__dict__)
            file.close()
            del tree
        except (IOError, EOFError, KeyError) as err:
            suffix = self.path + file_name if not 'dont save' in self.name else ''
            if pprint: print('>>>>> Calculating SOV expansion... ' + suffix)
            # set the computational tree
            self.setCompTree(eps=1.)
            # compute SOV factorisation
            self.calcSOVEquations(maxspace_freq=maxspace_freq, pprint=False)
            if not 'dont save' in self.name:
                # store the tree
                file = open(self.path + file_name, 'wb')
                dill.dump(self, file)
                file.close()


class CompartmentFitter(object):
    """
    Helper class to streamline fitting reduced compartmental models

    Attributes
    ----------
    tree: `neat.PhysTree()`
        The full tree based on which reductions are made
    name: str (default 'dont save')
        name of files in which intermediate trees required for the fit are
        stored. Can contain a path. Details about what is in the actual pickle
        files are appended as a suffix to `name`
    e_hs: np.array of float
        The holding potentials for which quasi active expansions are computed
    freqs: np.array of float or complex (default is ``np.array([0.])``)
        The frequencies at which impedance matrices are evaluated
    """

    def __init__(self, phys_tree,
                 e_hs=np.array([-75., -55., -35., -15.]), freqs=np.array([0.]),
                 name='dont save', path=''):
        self.tree = phys_tree.__copy__(new_tree=PhysTree())
        self.tree.treetype = 'original'
        # get all channels in the tree
        self.channel_names = self.tree.getChannelsInTree()
        # frequencies for fit
        self.freqs = freqs
        # expansion point holding potentials for fit
        self.e_hs = e_hs
        # name to store fit models
        self.name = name
        self.path = path

    def setCTree(self, loc_arg, extend_w_bifurc=True):
        """
        Store an initial `neat.CompartmentTree`, providing a tree
        structure scaffold for the fit for a given set of locations. The
        locations are also stored on ``self.tree`` under the name 'fit locs'

        Parameters
        ----------
        loc_arg: list of locations or string (see documentation of
                :func:`MorphTree._convertLocArgToLocs` for details)
            The compartment locations
        extend_w_bifurc: bool (optional, default `True`)
            To extend the compartment locations with all intermediate
            bifurcations (see documentation of
            :func:`MorphTree.extendWithBifurcationLocs`).
        """
        locs = self.tree._parseLocArg(loc_arg)
        if extend_w_bifurc:
            locs = self.tree.extendWithBifurcationLocs(locs)
        else:
            warnings.warn('Not adding bifurcations to `loc_arg`, this could '+ \
                          'lead to inaccurate fits. To add bifurcation, set' + \
                          'kwarg `extend_w_bifurc` to ``True``')
        self.tree.storeLocs(locs, name='fit locs')
        # create the reduced compartment tree
        self.ctree = self.tree.createCompartmentTree(locs)
        # # add currents to compartmental model
        for c_name, channel in self.tree.channel_storage.items():
            e_revs = []
            for node in self.tree:
                if c_name in node.currents:
                    e_revs.append(node.currents[c_name][1])
            # reversal potential is the same throughout the reduced model
            self.ctree.addCurrent(copy.deepcopy(channel), np.mean(e_revs))
        # set the equilibirum potentials at fit locations
        self.setEEq()

    def createTreeGF(self, channel_names=[]):
        """
        Create a `FitTreeGF` copy of the old tree, but only with the
        channels in ``channel_names``. Leak 'L' is included in the tree by
        default.

        Parameters
        ----------
        channel_names: list of strings
            List of channel names of the channels that are to be included in the
            new tree.

        Returns
        -------
        `FitTreeGF()`

        """
        # create new tree and empty channel storage
        tree = self.tree.__copy__(new_tree=FitTreeGF())
        tree.channel_storage = {}
        tree.setName(self.name, self.path)
        # add the ion channel to the tree
        channel_names_newtree = set()
        for node, node_orig in zip(tree, self.tree):
            node.currents = {}
            g_l, e_l = node_orig.currents['L']
            # add the current to the tree
            node._addCurrent('L', g_l, e_l)
            for channel_name in channel_names:
                try:
                    g_max, e_rev = node_orig.currents[channel_name]
                    node._addCurrent(channel_name, g_max, e_rev)
                    channel_names_newtree.add(channel_name)
                except KeyError:
                    pass

        tree.channel_storage = {channel_name: self.tree.channel_storage[channel_name] \
                                for channel_name in channel_names_newtree}
        tree.setCompTree()

        return tree

    def evalChannel(self, channel_name,
                          recompute=False, pprint=False, parallel=True, max_workers=None):
        """
        Evaluate the impedance matrix for the model restricted to a single ion
        channel type.

        Parameters
        ----------
        channel_name: string
            The name of the ion channel under consideration
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information
        parallel:  bool (optional, defaults to ``True``)
            whether the models are evaluated in parallel

        Return
        ------
        fit_mats
        """
        locs = self.tree.getLocs('fit locs')
        # find the expansion point parameters for the channel
        channel = self.tree.channel_storage[channel_name]
        sv_hs, e_hs = getExpansionPoints(self.e_hs, channel)
        n_tree = len(e_hs)

        # create the trees with only a single channel
        fit_tree = self.createTreeGF([channel_name])
        fit_tree.setName(self.name, self.path)
        fit_trees = []
        for sv_h, e_h in zip(sv_hs, e_hs):
            ftree = fit_tree.__copy__(new_tree=FitTreeGF())
            ftree.setExpansionPointsForFit({channel_name: sv_h}, e_h)
            fit_trees.append(ftree)

        # compute the impedance matrices for different activation levels
        args_list = [fit_trees,
                     [locs for _ in range(n_tree)],
                     [recompute for _ in range(n_tree)],
                     [pprint for _ in range(n_tree)]]
        # compute the impedance matrices
        if parallel:
            if max_workers is None:
                raise ValueError('need to provide number of workers if parallel is True')
            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as pool:
                fit_mats = list(pool.map(self._calcFitMatrices, *args_list))
        else:
            fit_mats = [self._calcFitMatrices(*args) for args in zip(*args_list)]

        # fit the model for this channel
        w_norm = 1. / np.sum([w_f for _, _, w_f in fit_mats])
        for _, _, w_f in fit_mats: w_f /= w_norm

        return fit_mats

    def _calcFitMatrices(self, fit_tree, locs, recompute, pprint):
        """
        Compute the matrices needed to fit the channel
        """
        e_h = fit_tree.root.e_eq
        c_name = list(fit_tree.channel_storage.keys())[0]
        sv_h = fit_tree.root.expansion_points[c_name]
        freqs = self.freqs
        # set the impedances in the tree
        fit_tree.setImpedancesInTree(recompute=recompute, pprint=pprint)
        # compute the impedance matrix for this acitvation level
        z_mat = fit_tree.calcImpedanceMatrix(locs)

        # compute the fit matrices
        m_f, v_t = self.ctree.computeGSingleChanFromImpedance(c_name, z_mat, e_h, freqs,
                        sv=sv_h, other_channel_names=['L'],
                        all_channel_names=self.channel_names, action='return')
        # compute open probability to weigh fit matrices
        channel = self.tree.channel_storage[c_name]
        po_h = channel.computePOpen(e_h, **sv_h)
        w_f = 1. / po_h

        return m_f, v_t, w_f

    def fitChannels(self, recompute=False, pprint=False, parallel=True):
        """
        Fit the active ion channel parameters

        Parameters
        ----------
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information
        parallel:  bool (optional, defaults to ``True``)
            whether the models are evaluated in parallel
        """
        # create the fit matrices for each channel
        n_arg = len(self.channel_names)

        if n_arg > 0:
            args_list = [self.channel_names,
                         [recompute for _ in range(n_arg)],
                         [pprint for _ in range(n_arg)],
                         [parallel for _ in range(n_arg)],
                        ]
            if parallel:
                max_workers = min(n_arg, cpu_count())
                # split cores evenly over inner workers
                inner_max_workers = cpu_count() // max_workers
                args_list += [[inner_max_workers for _ in range(n_arg)]]
                with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as pool:
                    fit_mats_ = list(pool.map(self.evalChannel, *args_list))
            else:
                fit_mats_ = [self.evalChannel(*args) for args in zip(*args_list)]
            fit_mats = [f_m for f_ms in fit_mats_ for f_m in f_ms]
            # store the fit matrices
            for m_f, v_t, w_f in fit_mats:
                if not (np.isnan(m_f).any() or np.isnan(v_t).any() or np.isnan(w_f).any()):
                    self.ctree._fitResAction('store', m_f, v_t, w_f,
                                             channel_names=self.channel_names)
            # run the fit
            self.ctree.runFit()

        # chan_eval = ChannelEvaluator()
        # chan_eval.evaluate(self, recompute=recompute, pprint=pprint, parallel=parallel)

    def fitPassive(self, use_all_channels=True, recompute=False, pprint=False):
        """
        Fit the steady state passive model, consisting only of leak and coupling
        conductances, but ensure that the coupling conductances takes the passive
        opening of all channels into account

        Parameters
        ----------
        use_all_channels: bool (optional)
            use leak at rest of all channels combined in the passive fit (passive
            leak has to be refit after capacitance fit)
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information

        """
        # get equilibirum potentials
        v_eqs_tree = self.getEEq('tree')
        v_eqs_fit = self.getEEq('fit')

        locs = self.tree.getLocs('fit locs')
        # initialize appropriate greens tree
        channel_names = list(self.tree.channel_storage.keys()) if use_all_channels \
                                                               else []
        fit_tree = self.createTreeGF(channel_names)
        fit_tree.setEEq(v_eqs_tree)
        fit_tree.setName(self.name + '_atRest_', self.path)
        # set the channels to passive
        fit_tree.asPassiveMembrane()
        # set the impedances in the tree
        fit_tree.setImpedancesInTree(recompute=recompute, pprint=pprint)
        # compute the steady state impedance matrix
        z_mat = fit_tree.calcImpedanceMatrix(locs)[0].real
        # fit the coupling+leak conductances to steady state impedance matrix
        self.ctree.computeGMC(z_mat, channel_names=['L'])

        # print passive impedance matrices
        if pprint:
            z_mat_fit = self.ctree.calcImpedanceMatrix(channel_names=['L'])
            np.set_printoptions(precision=2, edgeitems=10, linewidth=500, suppress=True)
            print('\n----- Impedance matrix comparison -----')
            print('> Zmat orig =')
            print(z_mat)
            print('> Zmat fit  =')
            print(z_mat_fit)
            print('> Zmat diff =')
            print(z_mat - z_mat_fit)
            print('---------------------------------------\n')
            # restore defaults
            np.set_printoptions(precision=8, edgeitems=3, linewidth=75, suppress=False)

    def fitPassiveLeak(self, recompute=False, pprint=True):
        """
        Fit leak only. Coupling conductances have to have been fit already.

        Parameters
        ----------
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information
        """
        locs = self.tree.getLocs('fit locs')
        # compute the steady state impedance matrix
        fit_tree = self.createTreeGF([])
        fit_tree.setName(self.name + '_onlyL_', self.path)
        # set the impedances in the tree
        fit_tree.setImpedancesInTree(recompute=recompute, pprint=pprint)
        # compute the steady state impedance matrix
        z_mat = fit_tree.calcImpedanceMatrix(locs)
        # fit the conductances to steady state impedance matrix
        self.ctree.computeGSingleChanFromImpedance('L', z_mat, -75., self.freqs,
                                                   other_channel_names=[],
                                                   action='fit')
        # print passive impedance matrices
        if pprint:
            z_mat_fit = self.ctree.calcImpedanceMatrix(channel_names=['L'])
            np.set_printoptions(precision=2, edgeitems=10, linewidth=500, suppress=True)
            print('\n----- Impedance matrix comparison -----')
            print('> Zmat orig =')
            print(z_mat)
            print('> Zmat fit  =')
            print(z_mat_fit)
            print('> Zmat diff =')
            print(z_mat - z_mat_fit)
            print('---------------------------------------\n')
            # restore defaults
            np.set_printoptions(precision=8, edgeitems=3, linewidth=75, suppress=False)

    def createTreeSOV(self, use_all_channels=False, eps=1.):
        """
        Create a `SOVTree` copy of the old tree

        Parameters
        ----------
        channel_names: list of strings
            List of channel names of the channels that are to be included in the
            new tree

        Returns
        -------
        `neat.PhysTree`

        """
        # create new tree and empty channel storage
        tree = self.tree.__copy__(new_tree=FitTreeSOV())
        if use_all_channels:
            tree.setName(self.name + '_allchans_', self.path)
        else:
            tree.setName(self.name, self.path)

        if not use_all_channels:
            tree.channel_storage = {}

            for node, node_orig in zip(tree, self.tree):
                node.currents = {}
                g_l, e_l = node_orig.currents['L']
                # add the current to the tree
                node._addCurrent('L', g_l, e_l)

        # set the computational tree
        tree.setCompTree(eps=eps)

        return tree

    def fitCapacitance(self, inds=[0], check_fit=True, force_tau_m_fit=False,
                             use_all_channels=True,
                             recompute=False, pprint=False, pplot=False):
        """
        Fit the capacitances of the model to the largest SOV time scale

        Parameters
        ----------
        inds: list of int (optional, defaults to ``[0]``)
            indices of eigenmodes used in the fit. Default is [0], indicating
            the largest eigenmode
        check_fit: bool (optional, default ``True``)
            Check whether the largest eigenmode of the reduced model is within
            tolerance of the largest eigenmode of the full tree. If not,
            capacitances are set to mach membrane time scale
        force_tau_m_fit: bool (optional, default ``False``)
            force capacitance fit through membrance time scale matching
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information
        pplot: bool (optional, defaults to ``False``)
            whether to plot the eigenmode timescales
        """
        locs = self.tree.getLocs('fit locs')
        # create an SOV tree
        sov_tree = self.createTreeSOV(use_all_channels=use_all_channels)
        # compute the SOV expansion for this tree
        sov_tree.setSOVInTree(recompute=recompute, pprint=pprint)
        # get SOV constants
        alphas, phimat, importance = sov_tree.getImportantModes(locarg=locs,
                                            sort_type='importance', eps=1e-12,
                                            return_importance=True)
        # fit the capacitances from SOV time-scales
        self.ctree.computeC(-alphas[inds]*1e3, phimat[inds,:],
                            weights=importance[inds])

        # fit capacitances vector fit
        # # greens_tree = self.createTreeGF()
        # channel_names = list(self.tree.channel_storage.keys())
        # greens_tree = self.createTreeGF(channel_names)
        # v_eqs_tree = self.getEEq('tree')
        # greens_tree.setEEq(v_eqs_tree)
        # greens_tree.asPassiveMembrane()
        # greens_tree.setName(self.name + '_atRest_')

        # greens_tree.setImpedancesInTree(many_freqs=True, recompute=False)
        # z_mat = greens_tree.calcImpedanceMatrix(locs)
        # # run the vector fit
        # self.ctree.computeCVF(greens_tree.freqs, z_mat)

        def calcTau():
            nm = len(locs)
            # original timescales
            taus_orig = np.sort(np.abs(1./alphas))[::-1][:nm]
            # fitted timescales
            lambdas, _, _ = self.ctree.calcEigenvalues()
            taus_fit = np.sort(np.abs(1./lambdas))[::-1]

            return taus_orig, taus_fit

        def calcTauM():
            clocs = [locs[n.loc_ind] for n in self.ctree]
            # original membrane time scales
            taus_m = []
            for l in clocs:
                g_m = sov_tree[l[0]].getGTot(channel_storage=sov_tree.channel_storage)
                taus_m.append(self.tree[l[0]].c_m / g_m *1e3)
            taus_m_orig = np.array(taus_m)
            # fitted membrance time scales
            taus_m_fit = np.array([node.ca / node.currents['L'][0]
                                   for node in self.ctree]) *1e3

            return taus_m_orig, taus_m_fit

        taus_orig, taus_fit = calcTau()
        if (check_fit and np.abs(taus_fit[0] - taus_orig[0]) > .8*taus_orig[0]) or \
           force_tau_m_fit:

            taus_m_orig, taus_m_fit = calcTauM()
            # if fit was not sane, revert to more basic membrane timescale match
            for ii, node in enumerate(self.ctree):
                node.ca = node.currents['L'][0] / taus_m_orig[ii]

            warnings.warn('No sane capacitance fit achieved for this configuragion,' + \
                          'reverted to more basic membrane time scale matching.')

        if pprint:
            # mode time scales
            taus_orig, taus_fit = calcTau()
            # membrane time scales
            taus_m_orig, taus_m_fit = calcTauM()

            np.set_printoptions(precision=2, edgeitems=10, linewidth=500, suppress=False)
            print('\n----- capacitances -----')
            print(('Ca (uF) =\n' + str([nn.ca for nn in self.ctree])))
            print('\n----- Eigenmode time scales -----')
            print(('> Taus original (ms) =\n' + str(taus_orig)))
            print(('> Taus fitted (ms) =\n' + str(taus_fit)))
            print('\n----- Membrane time scales -----')
            print(('> Tau membrane original (ms) =\n' + str(taus_m_orig)))
            print(('> Tau membrane fitted (ms) =\n' + str(taus_m_fit)))
            print('---------------------------------\n')
            # restore default print options
            np.set_printoptions(precision=8, edgeitems=3, linewidth=75, suppress=False)

        else:
            lambdas = None

        if pplot:
            lambdas, _, _ = self.ctree.calcEigenvalues()
            self.plotSOV(alphas, phimat, importance, n_mode=len(lambdas), alphas2=lambdas)
            self.plotKernels(alphas, phimat)
            # pl.show()

    def plotSOV(self, alphas, phimat, importance, n_mode=8, alphas2=None):
        fit_locs = self.tree.getLocs('fit locs')
        colours = list(pl.rcParams['axes.prop_cycle'].by_key()['color'])
        loc_colours = np.array([colours[ii%len(colours)] for ii in range(len(fit_locs))])
        markers = Line2D.filled_markers

        pl.figure('SOV', figsize=(10,10))
        gs = GridSpec(2,2)
        ax1, ax2, ax3 = pl.subplot(gs[0,0]), pl.subplot(gs[0,1]), pl.subplot(gs[1,:])
        # x axis modes
        x_arr = np.arange(n_mode)
        x_loc = np.arange(len(fit_locs))
        # time scales
        ax1.semilogy(x_arr, np.abs(1./alphas[:n_mode]), 'rD--')
        if alphas2 is not None:
            ax1.semilogy(x_arr[:len(alphas2)], np.sort(np.abs(1./alphas2))[::-1], 'bo--')
        ax1.set_xlabel(r'$k$')
        ax2.set_ylabel(r'$\tau_k$ (ms)')
        # importance
        ax2.semilogy(x_arr, importance[:n_mode], 'rD--')
        ax2.set_xlabel(r'$k$')
        ax2.set_ylabel(r'$I_k$')
        # spatial modes
        for kk in range(n_mode):
            ax3.plot(x_loc, phimat[kk,:], ls='--', c='DarkGrey')
            ax3.scatter(x_loc, phimat[kk,:], c=loc_colours, marker=markers[kk%len(markers)], label=r''+str(kk))
        ax3.set_xlabel(r'$x_i$')
        ax3.set_ylabel(r'$\phi_k(x_i)$')
        ax3.legend(loc=0)

    def _getKernels(self, a, c):
        nn = len(self.tree.getLocs('fit locs'))
        return [[Kernel((a, c[:,ii,jj])) for ii in range(nn)] for jj in range(nn)]

    def plotKernels(self, alphas, phimat, t_arr=None):
        fit_locs = self.tree.getLocs('fit locs')
        nn = len(fit_locs)

        if t_arr is None:
            t_arr = np.linspace(0.,200.,1e3)

        # compute eigenvalues
        alphas_comp, phimat_comp, phimat_inv_comp = \
                                self.ctree.calcEigenvalues(indexing='locs')
        # get the kernels
        k_orig = self._getKernels(alphas, np.einsum('ik,kj->kij', phimat.T, phimat))
        k_comp = self._getKernels(-alphas_comp, np.einsum('ik,kj->kij', phimat_comp, phimat_inv_comp))

        pl.figure('Kernels', figsize=(2.*nn, 1.5*nn))
        gs = GridSpec(nn, nn)
        gs.update(top=0.98, bottom=0.04, left=0.04, right=0.98)
        colours = list(pl.rcParams['axes.prop_cycle'].by_key()['color'])
        loc_colours = np.array([colours[ii%len(colours)] for ii in range(len(fit_locs))])

        for ii in range(nn):
            for jj in range(ii, nn):
                ko, kc = k_orig[ii][jj], k_comp[ii][jj]
                ax = pl.subplot(gs[ii,jj])
                ax.plot(t_arr, ko(t_arr), c='DarkGrey')
                ax.plot(t_arr, kc(t_arr), ls='--', c=loc_colours[jj])
                # limits
                ax.set_ylim((-0.5, 20.))
                # kernel label
                pstring = '%d $\leftrightarrow$ %d'%(ii,jj)
                ax.set_title(pstring, pad=-10)

    def checkPassive(self, locs, alpha_inds=[0], recompute=False, n_modes=5):
        self.setFitLocs(locs)
        # fit the passive steady state model
        self.fitPassive(recompute=recompute, pprint=True)
        # fit the capacitances
        self.fitCapacitance(inds=alpha_inds, recompute=recompute, pprint=True, pplot=True)

        fit_locs = self.tree.getLocs('fit locs')
        colours = list(pl.rcParams['axes.prop_cycle'].by_key()['color'])
        loc_colours = np.array([colours[ii%len(colours)] for ii in range(len(fit_locs))])

        pl.figure('tree')
        ax = pl.gca()
        locargs = [dict(marker='o', mec='k', mfc=lc, markersize=6.) for lc in loc_colours]
        self.tree.plot2DMorphology(ax, marklocs=fit_locs, locargs=locargs, use_radius=False)

        pl.tight_layout()
        pl.show()

    def getNET(self, c_loc, locs, channel_names=[], recompute=False, pprint=False):
        greens_tree = self.createTreeGF(channel_names=channel_names)
        greens_tree.setImpedancesInTree(recompute=recompute, pprint=False)
        # create the NET
        net, z_mat = greens_tree.calcNETSteadyState(c_loc)
        net.improveInputImpedance(z_mat)

        # prune the NET to only retain ``locs``
        loc_inds = greens_tree.getNearestLocinds([c_loc]+locs, 'net eval')
        net_reduced = net.getReducedTree(loc_inds, indexing='locs')

        return net_reduced

    def calcEEq(self, locs, t_max=500., dt=0.1, factor_lambda=10.):
        # create a biophysical simulation model
        sim_tree_biophys = self.tree.__copy__(new_tree=neurm.NeuronSimTree())
        # compute equilibrium potentials
        sim_tree_biophys.initModel(dt=dt, factor_lambda=factor_lambda)
        sim_tree_biophys.storeLocs(locs, 'rec locs', warn=False)
        res_biophys = sim_tree_biophys.run(t_max)
        sim_tree_biophys.deleteModel()
        return np.array([v_m[-1] for v_m in res_biophys['v_m']])

    def setEEq(self, t_max=500., dt=0.1, factor_lambda=10.):
        """
        Set equilibrium potentials, measured from neuron simulation. Sets the
        `v_eqs_tree` and `v_eqs_fit` attributes, respectively containing the
        equilibrium potentials at (the middle of) each node in the original
        tree and at each of the fit locations

        Parameters
        ----------
        t_max: float
            duration of the neuron simulation
        dt: float
            time-step of the neuron simulation
        factor_lambda: int of float
            if int, signifies the number of segments per section. If float,
            multiplies the number of segments given by the lambda rule with this
            number
        """
        tree_locs = [MorphLoc((n.index, .5), self.tree) for n in self.tree]
        fit_locs = self.tree.getLocs('fit locs')
        # compute equilibrium potentials
        v_eqs = self.calcEEq(tree_locs + fit_locs,
                             t_max=t_max, dt=dt, factor_lambda=factor_lambda)
        # store the equilibrium potentials
        self.v_eqs_tree = {n.index: v for n, v in zip(self.tree, v_eqs)}
        self.v_eqs_fit = v_eqs[len(tree_locs):]

    def getEEq(self, e_eqs_type, **kwargs):
        """
        Get equilibrium potentials. Specify
        `v_eqs_tree` and `v_eqs_fit` attributes, respectively containing the
        equilibrium potentials at (the middle of) each node in the original
        tree and at each of the fit locations

        Parameters
        ----------
        e_eqs_type: 'tree' or 'fit'
            For 'tree', returns the `v_eqs_tree` attribute, containing the
            equilibrium potentials at (the middle of) each node in the original
            tree. For 'fit', returns the `v_eqs_fit` attribute, containing the
            equilibrium potentials at each of the fit locations.
        kwargs: When `v_eqs_tree` or `v_eqs_fit`, have not been set, calls
            ::func::`self.setEEq()` with these `kwargs`

        """
        if not hasattr(self, 'v_eqs_tree') or not hasattr(self, 'v_eqs_fit'):
            self.setEEq(**kwargs)
        if e_eqs_type == 'fit':
            return self.v_eqs_fit
        elif e_eqs_type == 'tree':
            return self.v_eqs_tree
        else:
            raise IOError('``e_eqs_type`` should be \'fit\' or \'tree\'')

    def fitEEq(self, **kwargs):
        """
        Fits the leak potentials of the reduced model to yield the same
        equilibrium potentials as the full model

        Parameters
        ----------
        kwargs: When `v_eqs_tree` or `v_eqs_fit`, have not been set, calls
            ::func::`self.setEEq()` with these `kwargs`
        """
        # compute equilibirum potentials
        v_eqs = self.getEEq('fit', **kwargs)
        # fit the equilibirum potentials of the reduced model
        self.ctree.setEEq(v_eqs)
        self.ctree.fitEL()

    def fitModel(self, loc_arg, alpha_inds=[0], use_all_chans_for_passive=True,
                       recompute=False, pprint=False, parallel=False):
        """
        Runs the full fit for a set of locations (the location are automatically
        extended with the bifurcation locs)

        Parameters
        ----------
        loc_arg: list of locations or string (see documentation of
                :func:`MorphTree._convertLocArgToLocs` for details)
            The compartment locations
        alpha_inds: list of ints (optional, default to ``[0]``)
            Indices of all mode time-scales to be included in the fit
        use_all_chans_for_passive: bool (optional, default ``True``)
            Uses all channels in the tree to compute coupling conductances
        recompute: bool (optional, defaults to ``False``)
            whether to force recomputing the impedances
        pprint:  bool (optional, defaults to ``False``)
            whether to print information
        parallel:  bool (optional, defaults to ``False``)
            whether the models are evaluated in parallel

        Returns
        -------
        `neat.CompartmentTree`
            The reduced tree containing the fitted parameters
        """
        self.setCTree(loc_arg)
        # fit the passive steady state model
        self.fitPassive(recompute=recompute, pprint=pprint,
                        use_all_channels=use_all_chans_for_passive)
        # fit the capacitances
        self.fitCapacitance(inds=alpha_inds, use_all_channels=use_all_chans_for_passive,
                            recompute=recompute, pprint=pprint, pplot=False)
        # refit with only leak
        if use_all_chans_for_passive:
            self.fitPassiveLeak(recompute=recompute, pprint=pprint)

        # fit the ion channel
        self.fitChannels(recompute=recompute, pprint=pprint, parallel=parallel)
        # fit the resting potentials
        self.fitEEq()

        return self.ctree

    def recalcImpedanceMatrix(self, locarg, g_syns,
                              channel_names=None, recompute=False):
        # process input
        locs = self.tree._parseLocArg(locarg)
        n_syn = len(locs)
        assert n_syn == len(g_syns)
        if n_syn == 0:
            return np.array([[]])
        if channel_names is None:
            channel_names = list(self.tree.channel_storage.keys())

        # compute equilibirum potentials
        all_locs = [(n.index, .5) for n in self.tree]
        e_eqs = self.calcEEq(all_locs + locs)
        # create a greenstree with equilibrium potentials at rest
        greens_tree = self.createTreeGF(channel_names=channel_names)
        greens_tree.setName(self.name + '_atRest_', self.path)
        for ii, node in enumerate(greens_tree):
            node.setEEq(e_eqs[ii])
        greens_tree.setImpedancesInTree(recompute=recompute, pprint=False)
        # compute the impedance matrix of the synapse locations
        z_mat = greens_tree.calcImpedanceMatrix(locs, explicit_method=False)[0].real

        # get the reversal potentials of the synapse locations
        n_all = len(self.tree)
        e_eqs = e_eqs[n_all:]

        # compute the ZG matrix
        gd_mat = np.diag(g_syns)
        zg_mat = np.dot(z_mat, gd_mat)
        z_mat_ = np.linalg.solve(np.eye(n_syn) + zg_mat, z_mat)

        return z_mat_

    def fitSynRescale(self, c_locarg, s_locarg, comp_inds, g_syns, e_revs,
                            fit_impedance=False, channel_names=None, recompute=False):
        """
        Computes the rescaled conductances when synapses are moved to compartment
        locations, assuming a given average conductance for each synapse.

        Parameters
        ----------
        c_locarg: list of locations or string (see documentation of
                  :func:`MorphTree._convertLocArgToLocs` for details)
            The compartment locations
        s_locarg: list of locations or string (see documentation of
                  :func:`MorphTree._convertLocArgToLocs` for details)
            The synapse locations
        comp_inds: list or numpy.array of ints
            for each location in [s_locarg], gives the index of the compartment
            location in [c_locarg] to which the synapse is assigned
        g_syns: list or numpy.array of floats
            The average conductances for each synapse
        e_revs: list or numpy.array of floats
            The reversal potential of each synapse
        fit_impdedance: bool (optional, default `False`)
            Whether to also use the reproduction of the rescaled impedance matrix
            as target.
        channel_names: list of str or `None` (default)
            List of ion channels to be included in impedance matrix calculation.
            `None` includes all ion channels
        recompute: bool (defaults is `False`)
            Whether or not to recompute the impedance tree for this channel
            configuration

        Returns
        -------
        g_resc: numpy.array of floats
            The rescale values for the synaptic weights
        """
        # process input
        c_locs = self.tree._parseLocArg(c_locarg)
        s_locs = self.tree._parseLocArg(s_locarg)
        n_comp, n_syn = len(c_locs), len(s_locs)
        assert n_syn == len(g_syns) and n_syn == len(e_revs)
        assert len(c_locs) > 0
        if n_syn == 0:
            return np.array([])
        if channel_names is None:
            channel_names = list(self.tree.channel_storage.keys())
        cs_locs = c_locs + s_locs
        cg_syns = np.concatenate((np.zeros(n_comp), np.array(g_syns)))
        comp_inds, g_syns, e_revs = np.array(comp_inds), np.array(g_syns), np.array(e_revs)

        # compute equilibirum potentials
        all_locs = [(n.index, .5) for n in self.tree]
        e_eqs = self.calcEEq(all_locs + cs_locs)
        # create a greenstree with equilibrium potentials at rest
        greens_tree = self.createTreeGF(channel_names=channel_names)
        greens_tree.setName(self.name + '_atRest_', self.path)
        for ii, node in enumerate(greens_tree):
            node.setEEq(e_eqs[ii])
        greens_tree.setImpedancesInTree(recompute=recompute, pprint=False)
        # compute the impedance matrix of the synapse locations
        z_mat = greens_tree.calcImpedanceMatrix(cs_locs, explicit_method=False)[0].real
        zc_mat = z_mat[:n_comp, :n_comp]

        # get the reversal potentials of the synapse locations
        n_all = len(self.tree)
        e_cs = e_eqs[n_all:n_all+n_comp]
        e_ss = e_eqs[-n_syn:]

        # compute the ZG matrix
        gd_mat = np.diag(cg_syns)
        zg_mat_ = np.dot(z_mat, gd_mat)
        zg_mat = np.linalg.solve(np.eye(n_comp+n_syn) + zg_mat_, zg_mat_)
        zg_mat = zg_mat[:n_comp,n_comp:]

        # create the compartment assignment matrix & syn index vector
        c_mat = np.array([comp_inds == cc for cc in range(n_comp)]).astype(int)
        s_inds = np.array([np.where(cc > 0)[0][0] for cc in c_mat.T])

        # compute the driving potential vectors
        es_vec = e_revs - e_ss
        ec_vec = e_revs - e_cs[s_inds]

        zc_mat = np.dot(zc_mat, c_mat)
        czg_mat = np.dot(c_mat.T, zg_mat)

        # create matrices for inverse fit
        a1_mat = np.einsum('ck,kn->cnk', zc_mat, np.diag(ec_vec))
        a2_mat = np.einsum('ck,kn->cnk', zc_mat, czg_mat*es_vec[None,:])
        b_mat = zg_mat * es_vec[None,:]

        # unravel first two indices
        a_mat = np.reshape(a1_mat-a2_mat, (n_syn*n_comp,-1))
        b_vec = np.reshape(b_mat, (n_syn*n_comp,))

        if fit_impedance:
            # fit based on impedance matrix
            zr_mat = np.linalg.solve(np.eye(n_comp+n_syn) + zg_mat_, z_mat)

            zr_mat = zr_mat[:n_comp,:n_comp]
            zc_mat = z_mat[:n_comp,:n_comp]

            # b matrix for fit
            b_mat = zc_mat - zr_mat
            # a tensor for fit
            zcc = np.dot(zc_mat, c_mat)
            czr = np.dot(c_mat.T, zr_mat)
            aa_mat = np.einsum('ik,kn->ink', zcc, czr)

            # unravel first two indices
            a_mat_ = np.reshape(aa_mat, (n_comp*n_comp,-1))
            b_vec_ = np.reshape(b_mat, (n_comp*n_comp,))

            # perfor mfit
            a_mat = np.concatenate((a_mat, a_mat_), axis=0)
            b_vec = np.concatenate((b_vec, b_vec_), axis=0)

        # compute rescaled synaptic conductances
        g_resc = np.linalg.lstsq(a_mat, b_vec, rcond=None)[0]

        b_arr = g_syns > 1e-9
        g_resc[np.logical_not(b_arr)] = 1.
        g_resc[b_arr] = g_resc[b_arr] / g_syns[b_arr]

        return g_resc

    def assignLocsToComps(self, c_locarg, s_locarg, fz=.8,
                                channel_names=None, recompute=False):
        """
        assumes the root node is in `c_locarg`
        """
        if channel_names is None:
            channel_names = list(self.tree.channel_storage.keys())

        # compute equilibirum potentials
        e_eqs = self.getEEq('tree')
        # create a greenstree with equilibrium potentials at rest
        greens_tree = self.createTreeGF(channel_names=channel_names)
        greens_tree.setName(self.name + '_atRest_', self.path)
        for ii, node in enumerate(greens_tree):
            node.setEEq(e_eqs[ii])
        greens_tree.setImpedancesInTree(recompute=recompute, pprint=False)

        # process input
        c_locs = self.tree._parseLocArg(c_locarg)
        s_locs = self.tree._parseLocArg(s_locarg)
        # find nodes corresponding to locs
        c_nodes = [self.tree[loc['node']] for loc in c_locs]
        s_nodes = [self.tree[loc['node']] for loc in s_locs]
        # compute input impedances
        c_zins = [greens_tree.calcZF(c_loc, c_loc)[0] for c_loc in c_locs]
        s_zins = [greens_tree.calcZF(s_loc, s_loc)[0] for s_loc in s_locs]
        # paths to root
        c_ptrs = [self.tree.pathToRoot(node) for node in c_nodes]
        s_ptrs = [self.tree.pathToRoot(node) for node in s_nodes]

        c_inds = []
        for s_node, s_path, s_loc, s_zin in zip(s_nodes, s_ptrs, s_locs, s_zins):
            z_diffs = []
            # check if there are compartment nodes before bifurcation nodes in up direction
            nn_inds = greens_tree.getNearestNeighbourLocinds(s_loc, c_locs)
            # print c_before_b
            c_ns = [c_nodes[ii] for ii in nn_inds]
            c_ps = [c_ptrs[ii] for ii in nn_inds]
            c_ls = [c_locs[ii] for ii in nn_inds]
            c_zs = [c_zins[ii] for ii in nn_inds]
            for c_node, c_path, c_loc, c_zin in zip(c_ns, c_ps, c_ls, c_zs):
                # find the common node as far from the root as possible
                s_p, c_p = s_path[::-1], c_path[::-1]
                kk = 0
                while kk < min(len(s_p), len(c_p)) and s_p[kk] == c_p[kk]:
                    p_node = s_p[kk]
                    kk += 1
                # distinguish cases for computing impedance different
                if p_node == s_node and p_node != c_node:
                    z_diffs.append(fz*np.abs(c_zin - s_zin))
                elif p_node == c_node and p_node != s_node:
                    z_diffs.append((1.-fz)*np.abs(s_zin - c_zin))
                elif p_node == c_node and p_node == s_node:
                    fz_ = fz if c_loc['x'] > s_loc['x'] else (1.-fz)
                    z_diffs.append(fz_*np.abs(s_zin-c_zin))
                else:
                    b_loc = (p_node.index, 1.)
                    b_z = greens_tree.calcZF(b_loc, b_loc)[0]
                    z_diffs.append((1.-fz)*(c_zin - b_z) + fz * (s_zin - b_z))
            # compartment node with minimal impedance difference
            ind_aux = np.argmin(z_diffs)
            c_inds.append(nn_inds[ind_aux])

        return c_inds




import sympy as sp
import numpy as np
import scipy.optimize as so

import os
import copy
import warnings

CONC_DICT = {'na': 10.,  # mM
             'k': 54.4,  # mM
             'ca': 1e-4,  # 1e-4
             }

TEMP_DEFAULT = 36.

E_ION_DICT = {'na': 50.,
              'k': -85.,
              'ca': 50.,
            }


class _func(object):
    def __init__(self, eval_func_aux, eval_func_vtrap, e_trap):
        self.eval_func_aux = eval_func_aux
        self.eval_func_vtrap = eval_func_vtrap
        self.e_trap = e_trap

    def __call__(self, *args):
        vv = args[0]
        if isinstance(vv, float):
            if np.abs(vv - self.e_trap) < 0.001:
                return self.eval_func_vtrap(*args)
            else:
                return self.eval_func_aux(*args)
        else:
            fv_return = np.zeros_like(vv)
            bool_vtrap = np.abs(vv - self.e_trap) < 0.0001
            inds_vtrap = np.where(bool_vtrap)
            args_ = [a[inds_vtrap] for a in args]
            fv_return[inds_vtrap] = self.eval_func_vtrap(*args_)
            inds = np.where(np.logical_not(bool_vtrap))
            args_ = [a[inds] for a in args]
            fv_return[inds] = self.eval_func_aux(*args_)
            return fv_return


def _insert_function_prefixes(string, prefix='np',
                              functions=['exp', 'sin', 'cos', 'tan', 'pi']):
    """
    Prefix all occurences in the input `string` of the functions in the
    `functions` list with the provided `prefix`.

    Parameters
    ----------
    string: string
        the input string
    prefix: string, optional
        the prefix that is put before each function. Defaults to `'np'`
    functions: list of strings, optional
        the list of functions that will be prefixed. Defaults to
        `['exp', 'sin', 'cos', 'tan', 'pi']`

    Returns
    -------
    string

    Examples
    --------
    >>> _insert_function_prefixes('5. * exp(0.) + 3. * cos(pi)')
    '5. * np.exp(0.) + 3. * np.cos(pi)'
    """
    for func_name in functions:
        numpy_string = ''
        while len(string) > 0:
            ind = string.find(func_name)
            if ind == -1:
                numpy_string += string
                string = ''
            else:
                numpy_string += string[0:ind] + prefix + '.' + func_name
                string = string[ind+len(func_name):]
        string = numpy_string
    return string


class SPDict(dict):
    """
    Dictionary that accepts both strings and similarly name sympy symbols as keys
    """
    def __getitem__(self, key):
        try:
            return super(SPDict, self).__getitem__(key)
        except KeyError:
            if isinstance(key, sp.Symbol):
                return super(SPDict, self).__getitem__(str(key))
            else:
                return super(SPDict, self).__getitem__(sp.symbols(key))


class CallDict(dict):
    """
    Callable dictionary, items are supposed to be callables
    that all accept an identical argument list
    """
    def __call__(self, *args):
        """
        Calls dictionary items (supposed to be callable)
        """
        return SPDict({str(k): f(*args) for k, f in self.items()})


class BroadcastFunc(object):
    def __init__(self, func):
        self.func = func

    def __call__(self, *args):
        """
        Broadcast the result of `self.func` call to shape of first `arg`
        """
        res = self.func(*args)
        if isinstance(args[0], np.ndarray):
            return np.broadcast_to(res, args[0].shape)
        else:
            return res


class IonChannel(object):
    """
    Base ion channel class that implements linearization and code generation for
    NEURON (.mod-files) and C++.

    Userdefined ion channels should inherit from this class and implement the
    `define()` function, where the specific attributes of the ion channel are set.

    The ion channel current is of the form

    .. math:: i_{chan} = \overline{g} \, p_o(x_1, ... , x_n) \, (e - v)

    where $p_o$ is the open probability defined as a function of a number of
    state variables. State variables evolve according to

    .. math:: \dot{x}_i = f_i(x_i, v, c_1, ..., c_k)

    with $c_1, ..., c_n$ the (optional) set of concentrations the ion channel
    depends on. There are two canonical ways to define $f_i$, either based on
    reaction rates :math:`\\alpha` and :math:`\\beta`:

    .. math:: \dot{x}_i = \\alpha_i(v) \, (1 - x_i) - \\beta_i{v} \, x_i,

    or based on an asymptotic value :math:`x_i^{\infty}`  and time-scale :math:`\\tau_i`

    .. math:: \dot{x}_i = \\frac{x_i^{\infty}(v) - x_i}{\tau_i(v)}.

    `IonChannel` accepts handles either description. For the former description,
    dicts `self.alpha` and `self.beta` must be defined with as keys the names
    of every state variable in the open probability. Similarly, for the latter
    description, dicts `self.tauinf` and `self.varinf` must be defined with as
    keys the name of every state variable.

    The user **must** define the attributes `p_open`, and either `alpha` and
    `beta` or `tauinf` and `varinf` in the `define()` function. The other
    attributes `ion`, `conc`, `q10`, `temp`, and `e` are optional.

    Parameters
    ----------
    p_open: str
        The open probability of the ion channel.
    alpha, beta: dict {str: str}
        dictionary of the rate function for each state variables. Keys must
        correspond to the name of every state variable in `p_open`, values must
        be formulas written as strings with `v` and possible ion as variabels
    tauinf, varinf: dict {str: str}
        state variable time scale and asymptotic activation level. Keys must
        correspond to the name of every state variable in `p_open`, values must
        be formulas written as strings with `v` and possible ion as variabels

    ion: str ('na', 'ca', 'k' or ''), optional
        The ion to which the ion channel is permeable
    conc: set of str (containing 'na', 'ca', 'k') or dict of {str: float}
        The concentrations the ion channel activation depends on. Can be a set
        of ions or a dict with the ions as keys and default values as float.
    q10: str, optional
        Temperature dependence of the state variable rate functions. May be a
        float or a string convertible to a sympy expression containing the
        `temp` parameter (temperature in ``[deg C]``). This factor divides the
        time-scales :math:`\tau_i(v)` of the ion channel. If not given, default
        is 1.
    temp: float, optional
        The temperature at which the ion channel is evaluated. Can be modified
        after initializiation by calling
        `IonChannel.setDefaultParams(temp=new_temperature)`. If not given, the
        evaluates `self.q10` at the default temperature of 36 degC.
    e: float, optional
        Reversal of the ion channel in ``[mV]``. functions that need it allow
        the default value to be overwritten with a keyword argument. If nothing
        is provided, will take a default reversal for `self.ion` (which is
        -85 mV for 'K', 50 mV for 'Na' and 50 mV for 'Ca'). If no ion is
        provided, errors will occur if functions that need `e` are called
        without specifying the value as a keyword argument.

    Examples
    --------
    >>> class Na_Ta(IonChannel):
    >>>     def define(self):
    >>>         # from (Colbert and Pan, 2002), Used in (Hay, 2011)
    >>>         self.ion = 'na'
    >>>         # concentrations the ion channel depends on
    >>>         self.conc = {}
    >>>         # define channel open probability
    >>>         self.p_open = 'h * m ** 3'
    >>>         # define activation functions
    >>>         self.alpha, self.beta = {}, {}
    >>>         self.alpha['m'] =  '0.182 * (v + 38.) / (1. - exp(-(v + 38.) / 6.))' # 1/ms
    >>>         self.beta['m']  = '-0.124 * (v + 38.) / (1. - exp( (v + 38.) / 6.))' # 1/ms
    >>>         self.alpha['h'] = '-0.015 * (v + 66.) / (1. - exp( (v + 66.) / 6.))' # 1/ms
    >>>         self.beta['h']  =  '0.015 * (v + 66.) / (1. - exp(-(v + 66.) / 6.))' # 1/ms
    >>>         # temperature factor for reaction rates
    >>>         self.q10 = '2.3^((temp - 23.)/10.)'
    """
    def __init__(self):
        """
        Will give an ``AttributeError`` if initialized as is. Should only be
        initialized from its' derived classes that implement specific ion
        channel types.
        """
        # define the channel based on user specified state variables and activations
        self.define()

        # ion that carries the channel current
        if not hasattr(self, 'ion'):
            self.ion = ''

        # temperature factor, if it exist
        if not hasattr(self, 'q10'):
            self.q10 = '1.'
        self.q10 = sp.sympify(self.q10, evaluate=False)
        # sympy temperature symbols
        assert len(self.q10.free_symbols) <= 1
        self.sp_t = list(self.q10.free_symbols)[0] if len(self.q10.free_symbols) > 0 else \
                    sp.symbols('temp')

        # the voltage variable
        self.sp_v = sp.symbols('v')

        # extract the state variables
        self.p_open = sp.sympify(self.p_open)
        self.statevars = self.p_open.free_symbols

        # construct the rate functions
        if 'alpha' in self.__dict__ and 'beta' in self.__dict__:
            for svar in self.statevars:
                key = str(svar)
                self.alpha[svar] = sp.sympify(self.alpha[key], evaluate=False)
                self.beta[svar] = sp.sympify(self.beta[key], evaluate=False)
            # convert to internal representation
            self.varinf, self.tauinf = {}, {}
            for svar in self.statevars:
                key = str(svar)
                self.varinf[svar] = self.alpha[svar] / (self.alpha[svar] + self.beta[svar])
                self.tauinf[svar] = (1./self.q10) / (self.alpha[svar] + self.beta[svar])
        elif 'tauinf' in self.__dict__ and 'varinf' in self.__dict__:
            for svar in self.statevars:
                key = str(svar)
                self.varinf[svar] = sp.sympify(self.varinf[key], evaluate=False)
                self.tauinf[svar] = sp.sympify(self.tauinf[key], evaluate=False) / self.q10
                del self.varinf[key]
                del self.tauinf[key]
        else:
            raise AttributeError("Necessary attributes not defined, define either " + \
                                 "`alpha` and `beta` or `tauinf` and `varinf`.")
        self.varinf, self.tauinf = SPDict(self.varinf), SPDict(self.tauinf)

        # set the right hand side of the differential equation for
        # state variables
        self.fstatevar = SPDict()
        for svar in self.statevars:
            self.fstatevar[svar] = (-svar + self.varinf[svar]) / self.tauinf[svar]

        # concentrations the ion channel depends on
        if not hasattr(self, 'conc'):
            # if concentration ions are not defined, attempt to extract them from
            # the state variable functions
            self.conc = set()
            for key, expr in self.fstatevar.items():
                self.conc |= expr.free_symbols # set union
            # remove everything that is not a concentration
            self.conc -= self.statevars
            self.conc -= {self.sp_v, self.sp_t}
        # if no default concentrations are defined, default values are taken
        # from default concentration values
        if not hasattr(self.conc, 'values'):
            self.conc = SPDict({sp.symbols(str(ion)): \
                                CONC_DICT[str(ion)] for ion in self.conc})
        # sympy concentration symbols
        self.sp_c = [ion for ion in self.conc]

        # default parameters
        self.default_params = SPDict({})
        self.default_params[str(self.sp_t)] = self.temp if 'temp' in self.__dict__ else \
                                              TEMP_DEFAULT
        try:
            self.default_params['e'] = self.e if 'e' in self.__dict__ else \
                                       E_ION_DICT[self.ion]
        except KeyError:
            warnings.warn('No default reversal potential defined.')

        # set the lambda functions for efficient numpy evaluation
        self._lambdifyChannel()

    def __getstate__(self):
        """
        remove lambdified functions from dict as they can not be pickled
        """
        d = dict(self.__dict__)

        del d['f_statevar']
        del d['f_varinf']
        del d['f_tauinf']
        del d['f_p_open']
        del d['dp_dx'], d['df_dv'], d['df_dx'], d['df_dc']

        return d

    def __setstate__(self, s):
        """
        since lambdified functions were not pickled we need to restore them
        """
        self.__dict__ = s
        self._lambdifyChannel()

    def setDefaultParams(self, **kwargs):
        """
        **kwargs
            Default values for temperature (`temp`), reversal (`e`)
        """
        self.default_params.update(kwargs)
        self._lambdifyChannel()

    def _substituteDefaults(self, expr):
        """
        Substitute default values in input expression

        Parameters
        ----------
        expr: sympy expression
        """
        for param, val in self.default_params.items():
            expr = expr.subs(sp.symbols(param), val)
        return expr

    @property
    def ordered_statevars(self):
        return list(sorted(self.statevars, key=str))

    def _lambdifyChannel(self):
        """
        Create lambda functions based on sympy expression for relevant ion
        channel functions
        """
        # arguments for lambda function
        args = [self.sp_v] + self.ordered_statevars + self.sp_c
        args_ = [self.sp_v] + self.sp_c

        # lambdified open probability
        self.f_p_open = BroadcastFunc(sp.lambdify(args, self.p_open))
        # storatestate variable function
        self.f_statevar = CallDict()
        self.f_varinf, self.f_tauinf = CallDict(), CallDict()
        # storage of derivatives
        self.dp_dx = CallDict()
        self.df_dv, self.df_dx, self.df_dc = CallDict(), CallDict(), CallDict()

        for svar, f_svar in self.fstatevar.items():
            f_svar = self._substituteDefaults(f_svar)
            varinf = self._substituteDefaults(self.varinf[svar])
            tauinf = self._substituteDefaults(self.tauinf[svar])

            # state variable function
            self.f_statevar = BroadcastFunc(sp.lambdify(args, f_svar))

            # state variable activation & timescale
            self.f_varinf[svar] = BroadcastFunc(sp.lambdify(args_, varinf))
            self.f_tauinf[svar] = BroadcastFunc(sp.lambdify(args_, tauinf))

            # derivatives of open probability to state variables
            self.dp_dx[svar] = BroadcastFunc(
                                sp.lambdify(args, sp.diff(self.p_open, svar, 1)))

            # derivatives of state variable function to voltage
            self.df_dv[svar] = BroadcastFunc(
                                sp.lambdify(args, sp.diff(f_svar, self.sp_v, 1)))

            # derivatives of state variable function to state variable
            self.df_dx[svar] = BroadcastFunc(
                                sp.lambdify(args, sp.diff(f_svar, svar, 1)))

            # derivatives of state variable function to concentrations
            self.df_dc[svar] = \
                CallDict({c: BroadcastFunc(
                                sp.lambdify(args, sp.diff(f_svar, c, 1))) \
                          for c in self.sp_c})

    def _argsAsList(self, v, w_statevar=True, **kwargs):
        """
        Converts arguments to list for lambdified functions
        """
        arg_list = [v]

        if w_statevar:
            for svar in self.ordered_statevars:
                key = str(svar)
                try:
                    arg_list.append(kwargs[key])
                except KeyError:
                    # state variable is not in kwargs
                    # set default value based on voltage
                    args = self._argsAsList(v, w_statevar=False, **kwargs)
                    arg_list.append(self.f_varinf[svar](*args))

        for c in self.sp_c:
            key = str(c)
            try:
                arg_list.append(kwargs[key])
            except KeyError:
                # ion is not in kwargs
                # set stored default value
                arg_list.append(self.conc[c])

        return arg_list

    def computePOpen(self, v, **kwargs):
        """
        Compute the open probability of the ion channel

        Parameters
        ----------
        v: float or `np.ndarray` of float
            The voltage at which to evaluate the open probability
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        float or `np.ndarray` of float
            The open probability
        """
        args = self._argsAsList(v, **kwargs)
        return self.f_p_open(*args)

    def computeDerivatives(self, v, **kwargs):
        """
        Compute:
        (i) the derivatives of the open probability to the state variables
        (ii) The derivatives of state functions to the voltage
        (iii) The derivatives of state functions to the state variables

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage at which to evaluate the open probability
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        tuple of three floats or three `np.ndarray`s of float
            The derivatives
        """
        args = self._argsAsList(v, **kwargs)
        return self.dp_dx(*args), self.df_dv(*args), self.df_dx(*args)

    def computeDerivativesConc(self, v, **kwargs):
        """
        Compute the derivatives of the state functions to the concentrations

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage at which to evaluate the open probability
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        tuple of three floats or three `np.ndarray`s of float
            The derivatives
        """
        args = self._argsAsList(v, **kwargs)
        return self.df_dc(*args)

    def computeVarinf(self, v):
        """
        Compute the asymptotic values for the state variables at a given
        activation level

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage at which to evaluate the open probability

        Returns
        -------
        dict of `np.ndarray` of dict of float
            The asymptotic activations, items are of same type (and shape) as `v`
        """
        args = self._argsAsList(v, w_statevar=False, **{})
        return self.f_varinf(*args)

    def computeTauinf(self, v):
        """
        Compute the time-scales for the state variables at a given
        activation level

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage at which to evaluate the open probability

        Returns
        -------
        dict of `np.ndarray` of dict of float
            The asymptotic activations, items are of same type (and shape) as `v`
        """
        args = self._argsAsList(v, w_statevar=False, **{})
        return self.f_tauinf(*args)

    def computeLinear(self, v, freqs, **kwargs):
        """
        Combute the contributions of the state variables to the linearized
        channel current

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage ``[mV]`` at which to evaluate the open probability
        freqs float, complex, or `np.ndarray` of float or complex:
            The frequencies ``[Hz]`` at which to evaluate the linearized contribution
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        float, complex or `np.ndarray` of float or complex
            The linearized current. Shape is dimension of `freqs` followed by
            the dimensions of `v`.
        """
        dp_dx, df_dv, df_dx = self.computeDerivatives(v, **kwargs)

        # broadcast frequencies
        if isinstance(freqs, np.ndarray) and isinstance(v, np.ndarray):
            freqs.reshape(freqs.shape+tuple([1]*v.ndim))

        lin_f = np.zeros_like(freqs)
        for svar, dp_dx_ in dp_dx.items():
            df_dv_ = df_dv[svar] * 1e3 # convert to 1 / s
            df_dx_ = df_dx[svar] * 1e3 # convert to 1 / s
            # add to the impedance contribution
            lin_f += dp_dx_ * df_dv_ / (freqs - df_dx_)
        return lin_f

    def computeLinearConc(self, v, freqs, ion, **kwargs):
        """
        Combute the contributions of the state variables to the linearized
        channel current

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage ``[mV]`` at which to evaluate the open probability
        freqs: float, complex, or `np.ndarray` of float or complex:
            The frequencies ``[Hz]`` at which to evaluate the linearized contribution
        ion: str
            The ion name for which to compute the linearized contribution
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        float, complex or `np.ndarray` of float or complex
            The linearized current. Shape is dimension of `freqs` followed by
            the dimensions of `v`.
        """
        dp_dx, df_dv, df_dx = self.computeDerivatives(v, **kwargs)
        df_dc = self.computeDerivativesConc(v, **kwargs)

        # broadcast frequencies
        if isinstance(freqs, np.ndarray) and isinstance(v, np.ndarray):
            freqs.reshape(freqs.shape+tuple([1]*v.ndim))

        lin_f = np.zeros_like(freqs)
        for svar, dp_dx_ in dp_dx.items():
            df_dc_ = df_dc[svar][ion] * 1e3 # convert to 1 / s
            df_dx_ = df_dx[svar]      * 1e3 # convert to 1 / s
            # add to the impedance contribution
            lin_f += dp_dx_ * df_dc_ / (freqs - df_dx_)
        return lin_f

    def _getReversal(self, e):
        if e is None:
            try:
                e = self.default_params['e']
            except KeyError:
                raise KeyError('No default reversal defined, provide value for `e`.')
        return e

    def computeLinSum(self, v, freqs, e=None, **kwargs):
        """
        Combute the linearized channel current contribution
        (without concentributions from the concentration - see `computeLinConc()`)

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage ``[mV]`` at which to evaluate the open probability
        freqs: float, complex, or `np.ndarray` of float or complex:
            The frequencies ``[Hz]`` at which to evaluate the linearized contribution
        e: float or `None`
            The reversal potential of the channel. Defaults to the value stored
            in `self.default_params['e']` if not provided.
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        float, complex or `np.ndarray` of float or complex
            The linearized current. Shape is dimension of `freqs` followed by
            the dimensions of `v`.
        """
        e = self._getReversal(e)
        return (e - v) * self.computeLinear(v, freqs, **kwargs) - \
               self.computePOpen(v, **kwargs)

    def computeLinConc(self, v, freqs, ion, e=None, **kwargs):
        """
        Combute the linearized channel current contribution from the concentrations

        Parameters
        ----------
        v: float or `np.ndarray`
            The voltage ``[mV]`` at which to evaluate the open probability
        freqs: float, complex, or `np.ndarray` of float or complex:
            The frequencies ``[Hz]`` at which to evaluate the linearized contribution
        ion: str
            The ion name for which to compute the linearized contribution
        e: float or `None`
            The reversal potential of the channel. Defaults to the value stored
            in `self.default_params['e']` if not provided.
        **kwargs
            Optional values for the state variables and concentrations.
            Broadcastable to `v` if provided

        Returns
        -------
        float, complex or `np.ndarray` of float or complex
            The linearized current. Shape is dimension of `freqs` followed by
            the dimensions of `v`.
        """
        e = self._getReversal(e)
        return (e - v) * self.computeLinearConc(v, freqs, ion, **kwargs)


    def writeModFile(self, path, g=0., e=0.):
        """
        Writes a modfile of the ion channel for simulations with neuron
        """
        cname =  self.__class__.__name__
        sv = [str(svar) for svar in self.statevars]
        cs = [str(conc) for conc in self.conc]

        modname = 'I' + cname + '.mod'
        fname = os.path.join(path, modname)

        file = open(fname, 'w')

        file.write(': This mod file is automaticaly generated by the ' +
                    '``neat.channels.ionchannels`` module\n\n')

        file.write('NEURON {\n')
        file.write('    SUFFIX I%s\n'%cname)
        if self.ion == '':
            file.write('    NONSPECIFIC_CURRENT i' + '\n')
        else:
            file.write('    USEION %s WRITE i%s\n'%(self.ion, self.ion))
        for c in cs:
            file.write('    USEION %s READ %si\n'%(c, c))
        file.write('    RANGE  g, e' + '\n')

        taustring = 'tau_' + ', tau_'.join(sv)
        varstring = '_inf, '.join(sv) + '_inf'
        file.write('    GLOBAL %s, %s\n'%(varstring, taustring))
        file.write('    THREADSAFE' + '\n')
        file.write('}\n\n')

        file.write('PARAMETER {\n')
        file.write('    g = ' + str(g*1e-6) + ' (S/cm2)' + '\n')
        file.write('    e = ' + str(e) + ' (mV)' + '\n')
        for ion in cs:
            file.write('    ' + ion + 'i (mM)' + '\n')
        file.write('    celsius (degC)\n')
        file.write('}\n\n')

        file.write('UNITS {\n')
        file.write('    (mA) = (milliamp)' + '\n')
        file.write('    (mV) = (millivolt)' + '\n')
        file.write('    (mM) = (milli/liter)' + '\n')
        file.write('}\n\n')

        file.write('ASSIGNED {\n')
        file.write('    i%s (mA/cm2)\n'%self.ion)
        for var in sv:
            file.write('    %s_inf      \n'%var)
            file.write('    tau_%s (ms) \n'%var)
        file.write('    v (mV)' + '\n')
        file.write('    %s (degC)\n'%(self.sp_t))
        file.write('}\n\n')

        file.write('STATE {\n')
        for var in sv:
            file.write('    %s\n'%var)
        file.write('}\n\n')

        calcstring = 'i%s = g * (%s) * (v - e)'%(self.ion, sp.printing.ccode(self.p_open))

        file.write('BREAKPOINT {\n')
        file.write('    SOLVE states METHOD cnexp' + '\n')
        file.write('    %s\n'%calcstring)
        file.write('}\n\n')

        concstring = 'i, '.join(cs)
        if len(cs) > 0:
            concstring = ', ' + concstring
            concstring += 'i'

        file.write('INITIAL {\n')
        file.write('    rates(v%s)\n'%concstring)
        for var in sv:
            file.write('    %s = %s_inf\n'%(var,var))
        file.write('}\n\n')

        file.write('DERIVATIVE states {\n')
        file.write('    rates(v%s)\n'%concstring)
        for var in sv:
            file.write('    %s\' = (%s_inf - %s) /  tau_%s \n'%(var,var,var,var))
        file.write('}\n\n')

        # substitution for common neuron names
        repl_pairs = [(str(c), str(c)+'i') for c in self.conc]

        file.write('PROCEDURE rates(v%s) {\n'%concstring)
        file.write('    %s = celsius\n'%str(self.sp_t))
        for var, svar in zip(sv, self.statevars):
            vi = sp.printing.ccode(self.varinf[svar])
            ti = sp.printing.ccode(self.tauinf[svar])
            for repl_pair in repl_pairs:
                vi = vi.replace(*repl_pair)
                ti = ti.replace(*repl_pair)
            file.write('    %s_inf = %s\n'%(var, vi))
            file.write('    tau_%s = %s\n'%(var, ti))
        file.write('}\n\n')

        file.close()

        return modname

    def writeCPPCode(self, path):
        """
        Concentration dependent ion channels get constant concentrations
        substituted for c++ simulation
        """
        c_name = self.__class__.__name__
        svs = [str(svar) for svar in self.statevars]
        # rewrite open probabilities
        p_open_m = self.p_open
        p_open_m_inf = self.p_open
        for svar in self.statevars:
            p_open_m = p_open_m.subs(svar, sp.symbols('m_' + str(svar)))
            p_open_m_inf = p_open_m_inf.subs(svar, sp.symbols('m_' + str(svar) + '_inf'))
        # substitue concentrations in expression
        def _substituteConc(expr, prefix='', suffix=''):
            for ion, conc in self.conc.items():
                expr = expr.subs(ion, sp.symbols(prefix + str(ion) + suffix))
            return expr

        # open header and cc files
        fcc = open(os.path.join(path, 'Ionchannels.cc'), 'a')
        fh = open(os.path.join(path, 'Ionchannels.h'), 'a')

        # define class and functions in header file
        fh.write('class %s: public IonChannel{\n'%c_name)
        fh.write('private:' + '\n')
        for svar in self.statevars:
            sv = sp.printing.ccode(svar)
            fh.write('    double m_%s;\n'%sv)
            fh.write('    double m_%s_inf, m_tau_%s;\n'%(sv, sv))
            fh.write('    double m_v_%s = 10000.;\n'%sv)
        fh.write('    double m_p_open_eq = 0.0, m_p_open = 0.0;\n')
        # hardcode default concentrations
        for ion, conc in self.conc.items():
            fh.write('    double m_%s = %.8f;\n'%(ion, conc))
        fh.write('public:' + '\n')
        fh.write('    void calcFunStatevar(double v) override;' + '\n')
        fh.write('    double calcPOpen() override;' + '\n')
        fh.write('    void setPOpen() override;' + '\n')
        fh.write('    void setPOpenEQ(double v) override;' + '\n')
        fh.write('    void advance(double dt) override;' + '\n')
        fh.write('    double getCond() override;' + '\n')
        fh.write('    double getCondNewton() override;' + '\n')
        fh.write('    double f(double v) override;' + '\n')
        fh.write('    double DfDv(double v) override;' + '\n')
        fh.write('    void setfNewtonConstant(double* vs, int v_size) override;' + '\n')
        fh.write('    double fNewton(double v) override;' + '\n')
        fh.write('    double DfDvNewton(double v) override;' + '\n')
        fh.write('};' + '\n')

        # function in cc file
        fcc.write('void %s::calcFunStatevar(double v){\n'%c_name)
        for svar in self.statevars:
            varinf = self._substituteDefaults(self.varinf[svar])
            tauinf = self._substituteDefaults(self.tauinf[svar])
            sv = str(svar)
            vi = sp.printing.ccode(_substituteConc(varinf, prefix='m_'))
            ti = sp.printing.ccode(_substituteConc(tauinf, prefix='m_'))
            fcc.write('    m_%s_inf = %s;\n'%(sv, vi))
            # if self.varinf.shape[1] == 2 and ind == (0,0):
            if sv == 'm':
                # instantaneous approximation possible if statevar is activation (denoted by 'm')
                fcc.write('    if(m_instantaneous)' + '\n')
                fcc.write('        m_tau_%s = %s;\n'%(sv, sp.printing.ccode(sp.Float(1e-5))))
                fcc.write('    else' + '\n')
                fcc.write('        m_tau_%s = %s;\n'%(sv, ti))
            else:
                fcc.write('    m_tau_%s = %s;\n'%(sv, ti))
        fcc.write('}\n')

        fcc.write('double %s::calcPOpen(){\n'%c_name)
        fcc.write('    return %s;\n'%sp.printing.ccode(p_open_m))
        fcc.write('}\n')

        fcc.write('void %s::setPOpen(){\n'%c_name)
        fcc.write('    m_p_open = calcPOpen();\n')
        fcc.write('}\n')

        fcc.write('void %s::setPOpenEQ(double v){\n'%c_name)
        fcc.write('    calcFunStatevar(v);\n')
        fcc.write('\n')
        for sv in svs:
            fcc.write('    m_%s = m_%s_inf;\n'%(sv, sv))
        fcc.write('    m_p_open_eq = %s;\n'%sp.printing.ccode(p_open_m_inf))
        fcc.write('}\n')

        fcc.write('void %s::advance(double dt){\n'%c_name)
        for sv in svs:
            fcc.write('    double p0_%s = exp(-dt / m_tau_%s);\n'%(sv, sv))
            fcc.write('    m_%s *= p0_%s ;\n'%(sv, sv))
            fcc.write('    m_%s += (1. - p0_%s) *  m_%s_inf;\n'%(sv, sv, sv))
        fcc.write('}\n')

        fcc.write('double %s::getCond(){\n'%c_name)
        fcc.write('    return m_g_bar * (m_p_open - m_p_open_eq);\n')
        fcc.write('}\n')

        fcc.write('double %s::getCondNewton(){\n'%c_name)
        fcc.write('    return m_g_bar;\n')
        fcc.write('}\n')

        # function for temporal integration
        fcc.write('double %s::f(double v){\n'%c_name)
        fcc.write('    return (m_e_rev - v);\n')
        fcc.write('}\n')

        fcc.write('double %s::DfDv(double v){\n'%c_name)
        fcc.write('    return -1.;\n')
        fcc.write('}\n')

        # set voltage values to evaluate at constant voltage during newton iteration
        fcc.write('void %s::setfNewtonConstant(double* vs, int v_size){\n'%c_name)
        fcc.write('    if(v_size != %d)'%len(self.statevars) + '\n')
        fcc.write('        cerr << "input arg [vs] has incorrect size, ' + \
                  'should have same size as number of channel state variables" << endl' + ';\n')
        for ii, svar in enumerate(self.ordered_statevars):
            fcc.write('    m_v_%s = vs[%d];\n'%(str(svar), ii))
        fcc.write('}\n')

        # functions for solving Newton iteration
        fcc.write('double %s::fNewton(double v){\n'%c_name)
        p_o = self.p_open
        for svar in self.statevars:
            sv = 'v_' + str(svar)
            # substitute voltage symbol in the activation
            vi = self._substituteDefaults(self.varinf[svar]).subs(self.sp_v, sp.symbols(sv))
            vi = _substituteConc(vi, prefix='m_')
            # assign dynamic or fixed voltage to the activation
            fcc.write('    double %s;\n'%(sv))
            fcc.write('    if(m_%s > 1000.){\n'%sv)
            fcc.write('        %s = v;\n'%(sv))
            fcc.write('    } else{\n')
            fcc.write('        %s = m_%s;\n'%(sv, sv))
            fcc.write('    }' + '\n')
            fcc.write('    double %s = %s;\n'%(str(svar), sp.printing.ccode(vi)))

        fcc.write('    return (m_e_rev - v) * (%s - m_p_open_eq);\n'%sp.printing.ccode(self.p_open))
        fcc.write('}\n')

        fcc.write('double %s::DfDvNewton(double v){\n'%c_name)
        # p_o = self.p_open
        # compute partial derivatives
        # dp_o = np.zeros_like(self.statevars)
        # for ind, varname in np.ndenumerate(self.statevars):
        #     dp_o[ind] = sp.diff(p_o, varname, 1)
        dp_o = {svar: sp.diff(self.p_open, svar, 1) for svar in self.statevars}
        # print derivatives
        for svar in self.statevars:
            sv = 'v_' + str(svar)
            v_var = sp.symbols(sv)
            # substitute voltage symbol in the activation
            vi = self._substituteDefaults(self.varinf[svar]).subs(self.sp_v, v_var)
            vi = _substituteConc(vi, prefix='m_')
            dvi_dv = sp.diff(vi, v_var, 1)
            # compute derivative
            fcc.write('    double %s;\n'%sv)
            fcc.write('    double d%s_dv;\n'%str(svar))
            fcc.write('    if(m_%s > 1000.){\n'%sv)
            fcc.write('        %s = v;\n'%sv)
            fcc.write('        d%s_dv = %s;\n'%(str(svar), sp.printing.ccode(dvi_dv)))
            fcc.write('    } else{\n')
            fcc.write('        %s = m_%s;\n'%(sv, sv))
            fcc.write('        d%s_dv = 0;\n'%str(svar))
            fcc.write('    }\n')
            fcc.write('    double %s = %s;\n'%(str(svar), sp.printing.ccode(vi)))

        expr_str = ' + '.join(['%s * d%s_dv'%(sp.printing.ccode(dp_o[svar]), str(svar)) \
                               for svar in self.statevars])

        fcc.write('    return -1. * (%s - m_p_open_eq) + (%s) * (m_e_rev - v);\n'%(sp.printing.ccode(self.p_open), expr_str))
        fcc.write('}\n')

        fh.write('\n')
        fcc.write('\n')

        fh.close()
        fcc.close()

